(*******************************************************)
(*                                                     *)
(*                  Contexts Theory.                   *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* File   : hoaspack.v                                 *)
(* Author : Ivan Scagnetto (scagnett@dimi.uniud.it),   *)
(*          Dipartimento di Matematica e Informatica,  *)
(*          University of Udine                        *)
(* Date   : July 1998                                  *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* This file is part of "Pi-calculus in (Co)Inductive  *)
(* Type Theory", a joint work with Furio Honsell and   *)
(* Marino Miculan (accepted for publication on TCS).   *)
(* If you are interested you can find more information *)
(* (and a gzipped PostScript version of the article)   *)
(* at the following URL:                               *)
(* http://www.dimi.uniud.it/~scagnett/pi-calculus.html *)
(*                                                     *)
(*******************************************************)

Require Export picalcth.

Section monotonicity.

(*******************************************************)
(* The next axiom states that if a name does not occur *)
(* in an applied context, then it cannot occur in the  *)
(* context itself.                                     *)
(*******************************************************)

Axiom proc_mono : (p:name->proc)(x,y:name)(notin x (p y)) -> (notin x (nu p)).

(*********************************************)
(* Same as above, for free and bound labels. *)
(*********************************************)

Axiom f_act_mono : (a:name->f_act)(x,y:name)(f_act_notin x (a y)) -> (f_act_notin_ho x a).

Axiom b_act_mono : (a:name->b_act)(x,y:name)(b_act_notin x (a y)) -> (b_act_notin_ho x a).

End monotonicity.

Section expansions.

(*****************************************************)
(* Expansion of processes contexts into higher order *)
(* process contexts applied to names.                *)
(*****************************************************)

Axiom ho_proc_exp : (p:name->proc)(x:name)
                    (Ex [q:name->name->proc](notin x (nu [y:name](nu (q y))))/\p=(q x)).

Axiom ho2_proc_exp : (p:name->name->proc)(x:name)
                     (Ex [q:name->name->name->proc](notin x (nu [y:name](nu [z:name](nu (q y z)))))/\p=(q x)).

(*******************************)
(* Same as above, for process. *)
(*******************************)

Lemma proc_exp : (p:proc)(x:name)(Ex [q:name->proc](notin x (nu q))/\p=(q x)).

Proof.

(Induction p; Intros).

(Exists  [_:name]nil; Split;
 [ Apply notin_nu; Intros; Apply notin_nil | Try Trivial ]).

Elim (H x); Intros.
Exists [_:name](bang (x0 _)); Split.
Apply notin_nu; Intros.
Apply notin_bang; Inversion H0.
Inversion H2; Auto.
Inversion H0.
Rewrite <- H2; Trivial.

Elim (H x); Intros.
Exists [_:name](tau_pref (x0 _)); Split.
Apply notin_nu; Intros.
Apply notin_tau; Inversion H0.
Inversion H2; Auto.
Inversion H0.
Rewrite <- H2; Trivial.

Elim (H x); Elim (H0 x); Intros.
Inversion_clear H1; Inversion_clear H2.
Exists [_:name](par (x1 _) (x0 _)); Split.
Apply notin_nu; Intros.
Apply notin_par; [Inversion H1; Auto | Inversion H3; Auto].
Rewrite <- H4; Rewrite <- H5; Trivial.

Elim (H x); Elim (H0 x); Intros.
Inversion_clear H1; Inversion_clear H2.
Exists [_:name](sum (x1 _) (x0 _)); Split.
Apply notin_nu; Intros.
Apply notin_sum; [Inversion H1; Auto | Inversion H3; Auto].
Rewrite <- H4; Rewrite <- H5; Trivial.

Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H0; Exists [_:name](nu (x0 _)); Split; [Assumption | Rewrite H2; Trivial].

Elim (H x); Intros.
Inversion_clear H0.
Elim (LEM_name x n); Elim (LEM_name x n0); Intros.
Rewrite <- H0; Rewrite <- H3; Exists [_:name](match _ _ (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_match; Auto.
Rewrite H2; Trivial.
Rewrite <- H3; Exists [_:name](match _ n0 (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_match; Auto.
Rewrite H2; Trivial.
Rewrite <- H0; Exists [_:name](match n _ (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_match; Auto.
Rewrite H2; Trivial.
Exists [_:name](match n n0 (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_match; Auto.
Rewrite H2; Trivial.

Elim (H x); Intros.
Inversion_clear H0.
Elim (LEM_name x n); Elim (LEM_name x n0); Intros.
Rewrite <- H0; Rewrite <- H3; Exists [_:name](mismatch _ _ (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_mismatch; Auto.
Rewrite H2; Trivial.
Rewrite <- H3; Exists [_:name](mismatch _ n0 (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_mismatch; Auto.
Rewrite H2; Trivial.
Rewrite <- H0; Exists [_:name](mismatch n _ (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_mismatch; Auto.
Rewrite H2; Trivial.
Exists [_:name](mismatch n n0 (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_mismatch; Auto.
Rewrite H2; Trivial.

Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H0; Elim (LEM_name x n); Intros.
Rewrite <- H0; Exists [_:name](in_pref _ (x0 _)); Split; [Apply notin_nu; Intros; Apply notin_in; Intros; [Assumption | Inversion H1; Cut (notin x (nu (x0 z))); [Intro | Auto]; Inversion_clear H7; Auto] | Rewrite H2; Trivial].
Exists [_:name](in_pref n (x0 _)); Split; [Apply notin_nu; Intros; Apply notin_in; Intros; [Auto | Inversion H1; Cut (notin x (nu (x0 z))); [Intro | Auto]; Inversion_clear H7; Auto] | Rewrite H2; Trivial].

Elim (H x); Intros.
Inversion_clear H0.
Elim (LEM_name x n); Elim (LEM_name x n0); Intros.
Rewrite <- H0; Rewrite <- H3; Exists [_:name](out_pref _ _ (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_out; Auto.
Rewrite H2; Trivial.
Rewrite <- H3; Exists [_:name](out_pref _ n0 (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_out; Auto.
Rewrite H2; Trivial.
Rewrite <- H0; Exists [_:name](out_pref n _ (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_out; Auto.
Rewrite H2; Trivial.
Exists [_:name](out_pref n n0 (x0 _)); Split.
Apply notin_nu; Intros; Inversion H1; Apply notin_out; Auto.
Rewrite H2; Trivial.

Qed.

(*************************************************)
(* Expansion of free labels into label contexts. *)
(*************************************************)

Lemma f_act_exp : (a:f_act)(x:name)(Ex [b:name->f_act](f_act_notin_ho x b)/\a=(b x)).

Proof.

Unfold f_act_notin_ho; Induction a; Intros.

Exists  [x:name]tau; Split; [Intros | Trivial].
Apply f_act_notin_tau.

(Elim (LEM_name n x); Elim (LEM_name n0 x); Intros).
(Exists  [x:name](Out x x); Split).
(Intros; Apply f_act_notin_Out; Auto).
(Rewrite -> H; Rewrite -> H0; Try Trivial).

(Exists  [x:name](Out x n0); Split).
(Intros; Apply f_act_notin_Out; Auto).
(Rewrite -> H0; Try Trivial).

(Exists  [x:name](Out n x); Split).
(Intros; Apply f_act_notin_Out; Auto).
(Rewrite -> H; Try Trivial).

(Exists  [x:name](Out n n0); Split; [(Intros; Apply f_act_notin_Out; Auto) | Trivial]).

Qed.

(**************************************************)
(* Expansion of bound labels into label contexts. *)
(**************************************************)

Lemma b_act_exp : (a:b_act)(x:name)(Ex [b:name->b_act](b_act_notin_ho x b)/\a=(b x)).

Proof.

Unfold b_act_notin_ho; Induction a; Intros.

(Elim (LEM_name n x); Intros).
(Exists  [x:name](In x); Split).
(Intros; Apply b_act_notin_In; Auto).
(Rewrite -> H; Try Trivial).

(Exists  [x:name](In n); Split).
(Intros; Apply b_act_notin_In; Auto).
Trivial.

(Elim (LEM_name n x); Intros).
(Exists  [x:name](bOut x); Split).
(Intros; Apply b_act_notin_bOut; Auto).
(Rewrite -> H; Try Trivial).

(Exists  [x:name](bOut n); Split).
(Intros; Apply b_act_notin_bOut; Auto).
Trivial.

Qed.

End expansions.

(**********************************************************************)
(* Leibniz equality is a congruence w.r.t. replacement of free names. *)
(**********************************************************************)

Section eq_congr.

Variable x,y:name.

Axiom ho_proc_ext: (p,q:name->name->proc)(x:name)
                   (notin x (nu [y:name](nu (p y)))) ->
                   (notin x (nu [y:name](nu (q y)))) ->
                   (p x)=(q x)->p=q.

Axiom proc_ext : (p,q:name->proc)(x:name)
                 (notin x (nu p))->(notin x (nu q)) ->
                 (p x)=(q x) -> p=q.

Lemma proc_spec : (p,q:name->proc)p=q -> (x:name)(p x)=(q x).

Proof.

Intros; Rewrite H; Trivial.

Qed.

Lemma proc_sub_congr : (p,q:name->proc)
                       (notin x (nu p)) -> (notin x (nu q)) -> (p x)=(q x) -> (p y)=(q y).

Proof.

Intros; Apply proc_spec; Apply proc_ext with x; Auto.

Qed.

Lemma ho_proc_sub_congr : (p,q:name->name->proc)
                          (notin x (nu [z:name](nu (p z)))) ->
                          (notin x (nu [z:name](nu (q z)))) -> (p x)=(q x) -> (p y)=(q y).

Proof.

Intros; Elim (unsat (par (nu [z:name](nu (p z))) (nu [z:name](nu (q z))))
        (cons x (cons y empty))); Intros.
Inversion_clear H; Inversion_clear H0; Inversion_clear H2;
 Inversion_clear H0; Inversion_clear H4; Inversion_clear H2;
 Inversion_clear H5; Inversion_clear H6; Clear H7.

Apply proc_ext with x0; Auto.

Change ([_:name](p _ x0) y)=([_:name](q _ x0) y); Apply proc_sub_congr;
 Try Apply notin_nu; Intros;
 [Cut (notin x (nu (p z))) |
  Cut (notin x (nu (q z))) |
  Apply proc_spec]; Auto; Intro; Inversion_clear H7; Auto.

Qed.

Axiom f_act_ext : (a,b:name->f_act)(x:name)
                  (f_act_notin_ho x a) -> (f_act_notin_ho x b) ->
                  (a x)=(b x) -> a=b.

Lemma f_act_spec : (a,b:name->f_act)a=b -> (x:name)(a x)=(b x).

Proof.

Intros; Rewrite H; Trivial.

Qed.

Lemma f_act_sub_congr : (a,b:name->f_act)
                        (f_act_notin_ho x a) -> (f_act_notin_ho x b) ->
                        (a x)=(b x) -> (a y)=(b y).

Proof.

Intros; Apply f_act_spec; Apply f_act_ext with x; Auto.

Qed.

Axiom b_act_ext : (a,b:name->b_act)(x:name)
                  (b_act_notin_ho x a) -> (b_act_notin_ho x b) ->
                  (a x)=(b x) -> a=b.

Lemma b_act_spec : (a,b:name->b_act)a=b -> (x:name)(a x)=(b x).

Proof.

Intros; Rewrite H; Trivial.

Qed.

Lemma b_act_sub_congr : (a,b:name->b_act)
                        (b_act_notin_ho x a) -> (b_act_notin_ho x b) ->
                        (a x)=(b x) -> (a y)=(b y).

Proof.

Intros; Apply b_act_spec; Apply b_act_ext with x; Auto.

Qed.

End eq_congr.
