(*******************************************************)
(*                                                     *)
(*  Proof of cross adequacy between the two encodings  *)
(*            of Strong Late Bisimilarity.             *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* File   : xadequacy.v                                *)
(* Author : Ivan Scagnetto (scagnett@dimi.uniud.it),   *)
(*          Dipartimento di Matematica e Informatica,  *)
(*          University of Udine                        *)
(* Date   : July 1998                                  *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* This file is part of "Pi-calculus in (Co)Inductive  *)
(* Type Theory", a joint work with Furio Honsell and   *)
(* Marino Miculan (accepted for publication on TCS).   *)
(* If you are interested you can find more information *)
(* (and a gzipped PostScript version of the article)   *)
(* at the following URL:                               *)
(* http://www.dimi.uniud.it/~scagnett/pi-calculus.html *)
(*                                                     *)
(*******************************************************)

Require Export picalcth.

Section picalc_xadeq.

(*****************************************************************)
(* Monotonicity of the operator of the Strong Late Bisimilarity. *)
(*****************************************************************)

Lemma Op_StBisim_monotone : (R1,R2: proc->proc->Prop)
                            (Inclus R1 R2)->(Inclus (Op_StBisim R1) (Op_StBisim R2)).

Proof.

Unfold Inclus.
Intros.
Inversion H0.
Apply op_sb.
Split.

Intro.
Cut (((p0:proc)(ftrans p1 a p0)->(Ex [q1:proc](ftrans p2 a q1)/\(R1 p0 q1)))
/\((q1:proc)
    (ftrans p2 a q1)->(Ex [p0:proc](ftrans p1 a p0)/\(
R1 p0 q1)))
).
Intro.Elim H2.Intros.Clear H2.Split.Intros.
Cut (
(Ex [q1:proc](ftrans p2 a q1)/\(
R1 p0 q1))
).
Intro.
Elim H5.Intros.Clear H5.
Exists x.
Split.Tauto.Elim H6.
Intros.Clear H6.
Auto.Auto.Intros.
Cut (
(Ex [p0:proc](ftrans p1 a p0)/\(
R1 p0 q1))
).
Intro.
Elim H5.Intros.Clear H5.
Exists x.
Split.Tauto.Elim H6.
Intros.Clear H6.
Auto.Auto.Elim H1.Intros.Auto.

Split.

Intro.
Cut (
((p0:name->proc)
  (btrans p1 (In x) p0)
   ->(Ex [q1:name->proc]
          (btrans p2 (In x) q1)
          /\((y:name)(R1 (p0 y) (q1 y)))))
/\((q1:name->proc)
    (btrans p2 (In x) q1)
     ->(Ex [p0:name->proc]
            (btrans p1 (In x) p0)
            /\((y:name)(R1 (p0 y) (q1 y)))))
).
Intro.Elim H2.Intros.Clear H2.Split.Intros.
Cut (
(Ex [q1:name->proc]
     (btrans p2 (In x) q1)/\((y:name)(R1 (p0 y) (q1 y))))
).
Intro.
Elim H5.Intros.Clear H5.
Exists x0.
Split.Tauto.Elim H6.
Intros.Clear H6.
Auto.Auto.Intros.
Cut (
(Ex [p0:name->proc]
     (btrans p1 (In x) p0)/\((y:name)(R1 (p0 y) (q1 y))))

).
Intro.
Elim H5.Intros.Clear H5.
Exists x0.
Split.Tauto.Elim H6.
Intros.Clear H6.
Auto.Auto.Elim H1.Intros.Elim H3.Intros.Auto.

Intro.
Cut (
((p0:name->proc)
  (btrans p1 (bOut x) p0)
   ->(Ex [q1:name->proc]
          (btrans p2 (bOut x) q1)
          /\((y:name)(notin y (nu p0)) -> (notin y (nu q1)) -> (R1 (p0 y) (q1 y)))))
/\((q1:name->proc)
    (btrans p2 (bOut x) q1)
     ->(Ex [p0:name->proc]
            (btrans p1 (bOut x) p0)
            /\((y:name)(notin y (nu p0)) -> (notin y (nu q1)) -> (R1 (p0 y) (q1 y)))))
).
Intros.Elim H2.Intros.Clear H2.Split.Intros.
Cut (
(Ex [q1:name->proc]
     (btrans p2 (bOut x) q1)/\((y:name)(notin y (nu p0)) -> (notin y (nu q1)) -> (R1 (p0 y) (q1 y))))
).
Intro.
Elim H5.Intros.Clear H5.
Exists x0.
Split.Tauto.Elim H6.
Intros.Clear H6.
Auto.Auto.Intros.
Cut (
(Ex [p0:name->proc]
     (btrans p1 (bOut x) p0)/\((y:name)(notin y (nu p0)) -> (notin y (nu q1)) -> (R1 (p0 y) (q1 y))))

).
Intro.
Elim H5.Intros.Clear H5.
Exists x0.
Split.Tauto.Elim H6.
Intros.Clear H6.
Auto.Auto.Elim H1.Intros.Elim H3.Intros.Auto.

Qed.

(****************************)
(* Internal Cross Adequacy. *)
(****************************)

Lemma Soundness : (p1,p2:proc)(StBisim p1 p2)->(StBisim' p1 p2).

Proof.

Intros.
Apply Co_Ind with StBisim .
Unfold Inclus .
Intros.
Inversion H0.
Apply op_sb.
Auto.Auto.

Qed.

Lemma Completeness : (p1,p2:proc)(StBisim' p1 p2)->(StBisim p1 p2).

Proof.

Intros.Elim H.Unfold Inclus.Do 2 Intro. Generalize p1 p2.
Cofix.Intros.Case (H0 p0 p3 H1).Intros.

Apply sb.Split.Intro.
Cut (
((p1:proc)(ftrans p0 a p1)->(Ex [q1:proc]((ftrans p3 a q1) /\ (R p1 q1))))
/\
((q1:proc)(ftrans p3 a q1)->(Ex [p1:proc]((ftrans p0 a p1) /\ (R p1 q1))))
    ).
Intro.Elim H3.Intros.Clear H3.Split.Intros.Cut(Ex [q1:proc](ftrans p3 a q1)/\(R p4 q1)).
Intro.Elim H6.Intros.Clear H6.Exists x.Split.Tauto.Apply Completeness.Tauto.Auto.Intros.
Cut (Ex [p1:proc](ftrans p0 a p1)/\(R p1 q1)).Intro.Elim H6.Intros.Clear H6.Exists x.Split.Tauto.
Apply Completeness.Tauto.Auto.Elim H2.Intros.Auto.

Split.
Intro.
Cut (
((p1:name->proc)
  (btrans p0 (In x) p1)
   ->(Ex [q1:name->proc]
          (btrans p3 (In x) q1)
          /\((y:name)(R (p1 y) (q1 y)))))
/\((q1:name->proc)
    (btrans p3 (In x) q1)
     ->(Ex [p1:name->proc]
            (btrans p0 (In x) p1)
            /\((y:name)(R (p1 y) (q1 y)))))
    ).
Intro.Elim H3.Intros.Clear H3.Split.Intros.
Cut (Ex [q1:name->proc]
         (btrans p3 (In x) q1)
         /\((y:name)(R (p4 y) (q1 y)))).
Intro.Elim H6.Intros.Clear H6.Exists x0.Split.Tauto.Intro.Apply Completeness.Elim H7.Intros.Auto.
Auto.Intros.
Cut (Ex [p1:name->proc]
         (btrans p0 (In x) p1)
         /\((y:name)(R (p1 y) (q1 y)))).
Intro.Elim H6.Intros.Clear H6.Exists x0.Split.Tauto.Intro.Apply Completeness.Elim H7.Intros.Auto.
Auto.Elim H2.Intros.Elim H4.Intros.Auto.

Intro.
Cut (
((p1:name->proc)
  (btrans p0 (bOut x) p1)
   ->(Ex [q1:name->proc]
          (btrans p3 (bOut x) q1)
          /\((y:name)(notin y (nu p1)) -> (notin y (nu q1)) -> (R (p1 y) (q1 y)))))
/\((q1:name->proc)
    (btrans p3 (bOut x) q1)
     ->(Ex [p1:name->proc]
            (btrans p0 (bOut x) p1)
            /\((y:name)(notin y (nu p1)) -> (notin y (nu q1)) -> (R (p1 y) (q1 y)))))
    ).
Intro.Elim H3.Intros.Clear H3.Split.Intros.
Cut (Ex [q1:name->proc]
         (btrans p3 (bOut x) q1)
         /\((y:name)(notin y (nu p4)) -> (notin y (nu q1)) -> (R (p4 y) (q1 y)))).
Intro.Elim H6.Intros.Clear H6.Exists x0.Split.Tauto.Intros.Apply Completeness.Elim H7.Intros.Auto.
Auto.Intros.
Cut (Ex [p1:name->proc]
         (btrans p0 (bOut x) p1)
         /\((y:name)(notin y (nu p1)) -> (notin y (nu q1)) -> (R (p1 y) (q1 y)))).
Intro.Elim H6.Intros.Clear H6.Exists x0.Split.Tauto.Intros.Apply Completeness.Elim H7.Intros.Auto.
Auto.Elim H2.Intros.Elim H4.Intros.Auto.

Qed.

End picalc_xadeq.
