(*******************************************************)
(*                                                     *)
(*         Formalization of the Algebraic Laws         *)
(*            of Strong Late Bisimilarity.             *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* File   : alglaws.v                                  *)
(* Author : Ivan Scagnetto (scagnett@dimi.uniud.it),   *)
(*          Dipartimento di Matematica e Informatica,  *)
(*          University of Udine                        *)
(* Date   : July 1998                                  *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* This file is part of "Pi-calculus in (Co)Inductive  *)
(* Type Theory", a joint work with Furio Honsell and   *)
(* Marino Miculan (accepted for publication on TCS).   *)
(* If you are interested you can find more information *)
(* (and a gzipped PostScript version of the article)   *)
(* at the following URL:                               *)
(* http://www.dimi.uniud.it/~scagnett/pi-calculus.html *)
(*                                                     *)
(*******************************************************)

Require Export basiclemmata.

Section equivalence.

Lemma REF: (p:proc)(StBisim p p).

Proof.

(* This lemma is proved from scratch: it is a coinductive proof. *)

Cofix; Intro; Apply sb; Do 3 Try (Split; Intros).
Exists p1; Split; Auto.
Exists q1; Split; Auto.
Exists p1; Split; [Assumption | Intro; Auto].
Exists q1; Split; [Assumption | Intro; Auto].
Exists p1; Split; [Assumption | Intros; Auto].
Exists q1; Split; [Assumption | Intros; Auto].

Qed.

Lemma SYM: (p,q:proc)(StBisim p q) -> (StBisim q p).

Proof.

(* This lemma is proved from scratch: it is a coinductive proof. *)

Cofix; Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion H.
Elim H1; Intros.
Elim (H4 a); Intros.
Elim (H7 p1 H0); Intros.
Elim H8; Intros; Exists x; Split; [Assumption | Apply SYM; Assumption].

Inversion H.
Elim H1; Intros.
Elim (H4 a); Intros.
Elim (H6 q1 H0); Intros.
Elim H8; Intros; Exists x; Split; [Assumption | Apply SYM; Assumption].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H6 x); Intros.
Elim (H9 p1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intro; Apply SYM; Auto].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H6 x); Intros.
Elim (H8 q1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intro; Apply SYM; Auto].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H7 x); Intros.
Elim (H9 p1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intros; Apply SYM; Apply H12; Auto].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H7 x); Intros.
Elim (H8 q1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intros; Apply SYM; Apply H12; Auto].

Qed.

Transparent SYM.

Lemma TRANS: (p,q,r:proc)(StBisim p q) -> (StBisim q r) -> (StBisim p r).

Proof.

(* This lemma depends on Lemma L6_Light: it is a coinductive proof. *)

Cofix.
Do 4 Intro.
Inversion_clear H.
Intro.
Inversion_clear H.
Apply sb.
Split.
Intro.
Cut ((p1:proc)
      (ftrans p a p1)->(Ex [q1:proc](ftrans q a q1)/\(StBisim p1 q1)))
    /\((q1:proc)
        (ftrans q a q1)->(Ex [p1:proc](ftrans p a p1)/\(StBisim p1 q1))).
Intro.
Cut ((p1:proc)
      (ftrans q a p1)->(Ex [q1:proc](ftrans r a q1)/\(StBisim p1 q1)))
    /\((q1:proc)
        (ftrans r a q1)->(Ex [p1:proc](ftrans q a p1)/\(StBisim p1 q1))).
Intro.
Inversion_clear H.
Inversion_clear H2.
Split.
Intros.
Cut (Ex [q1:proc](ftrans q a q1)/\(StBisim p1 q1)).
Intro.
Inversion_clear H6.
Cut (Ex [q1:proc](ftrans r a q1)/\(StBisim x q1)).
Intro.
Inversion_clear H6.
Exists  x0.
Split.
Tauto.

Apply TRANS with x.
Tauto.

Tauto.

Apply H.
Tauto.

Apply H3.
Assumption.

Intros.
Cut (Ex [p1:proc](ftrans q a p1)/\(StBisim p1 q1)).
Intro.
Inversion_clear H6.
Cut (Ex [p1:proc](ftrans p a p1)/\(StBisim p1 x)).
Intro.
Inversion_clear H6.
Exists  x0.
Split.
Tauto.

Apply TRANS with x.
Tauto.

Tauto.

Apply H4.
Tauto.

Apply H5.
Assumption.

Elim H1.
Intros.
Apply H2.

Elim H0.
Intros.
Apply H.

Split.
Intro.
Cut ((p1:name->proc)
      (btrans p (In x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans q (In x) q1)
            /\((y:name)(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans q (In x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans p (In x) p1)
              /\((y:name)(StBisim (p1 y) (q1 y))))).
Intro.
Cut ((p1:name->proc)
      (btrans q (In x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans r (In x) q1)
            /\((y:name)(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans r (In x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans q (In x) p1)
              /\((y:name)(StBisim (p1 y) (q1 y))))).
Intro.
Inversion_clear H.
Inversion_clear H2.
Split.
Intros.

Cut (Ex
      [q1:name->proc]
       (btrans q (In x) q1)
       /\((y:name)(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [q1:name->proc]
       (btrans r (In x) q1)
       /\((y:name)(StBisim (x0 y) (q1 y)))).
Intro.
Inversion_clear H6.
Exists  x1.
Split.
Tauto.

Intro.
Apply TRANS with (x0 y).
Generalize y .
Tauto.

Generalize y .
Tauto.

Apply H.
Tauto.

Apply H3.
Assumption.

Intros.
Cut (Ex
      [p1:name->proc]
       (btrans q (In x) p1)
       /\((y:name)(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [p1:name->proc]
       (btrans p (In x) p1)
       /\((y:name)(StBisim (p1 y) (x0 y)))).
Intro.
Inversion_clear H6.
Exists  x1.
Split.
Tauto.

Intro.
Apply TRANS with (x0 y).
Generalize y .
Tauto.

Generalize y .
Tauto.

Apply H4.
Tauto.

Apply H5.
Assumption.

Elim H1.
Intros.
Elim H3.
Intros.
Apply H4.

Elim H0.
Intros.
Elim H2.
Intros.
Apply H3.

Intro.
Cut ((p1:name->proc)
      (btrans p (bOut x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans q (bOut x) q1)
            /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans q (bOut x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans p (bOut x) p1)
              /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y))))).
Intro.
Cut ((p1:name->proc)
      (btrans q (bOut x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans r (bOut x) q1)
            /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans r (bOut x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans q (bOut x) p1)
              /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y))))).
Intro.
Inversion_clear H.
Inversion_clear H2.
Split.
Intros.
Cut (Ex
      [q1:name->proc]
       (btrans q (bOut x) q1)
       /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [q1:name->proc]
       (btrans r (bOut x) q1)
       /\((y:name)
                 (notin y (nu x0))
                  ->(notin y (nu q1))->(StBisim (x0 y) (q1 y)))).
Intro.
Inversion_clear H6.
Exists x1.
Split.
Tauto.

Intros.
(Elim (ho_proc_exp x0 y); Intros).
Inversion_clear H10.
Elim (unsat (par (nu p1) (par (nu x0) (par (nu x1) (nu [w:name](nu (x2 w)))))) (cons y empty)); Intros.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H14.
Clear H18.

Apply TRANS with (x2 x3 y).
Elim (unsat (par (nu p1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](p1 x4) x3) ([z:name](x2 z x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H6.
Apply H24.
Auto.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Inversion_clear H7.
(Rewrite <- H12; Apply H23).
Assumption.
Assumption.

Apply notin_nu.
Intros.
Inversion H10.
(Apply H24; Auto).

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Assumption.

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Elim (unsat (par (nu x1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](x2 z x4) x3) ([z:name](x1 x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H9.
Apply H24.
Auto.

Inversion_clear H8.
(Rewrite <- H12; Apply H23).
Assumption.
Assumption.

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H15.
(Apply H24; Auto).

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Assumption.

Apply H.Tauto.
Apply H3.Tauto.

Intros.
Cut (Ex
      [p1:name->proc]
       (btrans q (bOut x) p1)
       /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [p1:name->proc]
       (btrans p (bOut x) p1)
       /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu x0))->(StBisim (p1 y) (x0 y)))).
Intro.
Inversion_clear H6.
Exists  x1.
Split.
Tauto.

Intros.
(Elim (ho_proc_exp x0 y); Intros).
Inversion_clear H10.
Elim (unsat (par (nu x1) (par (nu x0) (par (nu q1) (nu [w:name](nu (x2 w)))))) (cons y empty)); Intros.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H14.
Clear H18.
Apply TRANS with (x2 x3 y).
Elim (unsat (par (nu x1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](x1 x4) x3) ([z:name](x2 z x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H6.
Apply H24.
Auto.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Inversion_clear H8.
(Rewrite <- H12; Apply H23).
Assumption.

Assumption.

Apply notin_nu.
Intros.
Inversion H10.
(Apply H24; Auto).

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Assumption.

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Elim (unsat (par (nu q1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](x2 z x4) x3) ([z:name](q1 x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H9.
Apply H24.
Auto.

Inversion_clear H7.
(Rewrite <- H12; Apply H23).
Assumption.

Assumption.

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H15.
(Apply H24; Auto).

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Assumption.

Apply H4.Tauto.
Apply H5.Tauto.

Elim H1.
Intros.
Elim H3.
Intros.
Apply H5.

Elim H0.
Intros.
Elim H2.
Intros.
Apply H4.

Qed.

Transparent TRANS.

End equivalence.

Section structural_congruence1.

Lemma TAU_S : (p,q:proc)(StBisim p q)->(StBisim (tau_pref p) (tau_pref q)).

Proof.

(* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).
Inversion H0; Exists  q; Split; 
   [ Apply TAU | Rewrite <- H3]; Assumption.
Inversion H0; Exists  p; Split;
   [ Apply TAU | Rewrite <- H3 ]; Assumption.
Inversion_clear H0.
Inversion_clear H0.
Inversion_clear H0.
Inversion_clear H0.

Qed.

Transparent TAU_S.

Lemma SUM_S: (p,q,r:proc)(StBisim p q)->(StBisim (sum p r) (sum q r)).

Proof.

(* This lemma depends on Lemma REF. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).

Inversion H0.
Inversion H.
Inversion_clear H6.
(Elim (H9 a); Intros).
(Elim (H6 p1 H5); Intros).
(Exists  x; Split; [ Apply fSUM1; Tauto | Tauto ]).
(Exists p1; Split; [Apply fSUM2; Tauto | Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
(Elim (H9 a); Intros).
(Elim (H11 q1 H5); Intros).
(Exists  x; Split; [ Apply fSUM1; Tauto | Tauto ]).
(Exists q1; Split; [Apply fSUM2; Tauto | Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H6 x); Intros).
(Elim (H10 p1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists p1; Split; [Apply bSUM2; Tauto | Intro; Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H6 x); Intros).
(Elim (H12 q1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists q1; Split; [Apply bSUM2; Tauto | Intro; Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H11 x); Intros).
(Elim (H10 p1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists p1; Split; [Apply bSUM2; Tauto | Intros; Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H11 x); Intros).
(Elim (H12 q1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists q1; Split; [Apply bSUM2; Tauto | Intros; Apply REF]).

Qed.

Transparent SUM_S.

Lemma NU_S: (p,q:name->proc)
                  ((x:name)(notin x (nu p))->(notin x (nu q))
                    ->(StBisim (p x) (q x)))
                      ->(StBisim (nu p) (nu q)).

Proof. 

(* This lemma depends on Lemmata FTR_L3, BTR_L3 and L6_Light: it is a coinductive proof. *)

(Cofix; Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Generalize H2.
Generalize H0.
Elim a.
Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H10.
Inversion_clear H12.
Elim (H10 tau); Intros.
Cut (ftrans (p x) tau (p2 x)); [Intro | Apply H6; Auto].
Elim (H12 (p2 x) H15); Intros.
Inversion_clear H16.
Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists (nu x1); Split.

Apply fRES with empty; Intros; Auto.
Clear H22 H23.
(Change (ftrans (q y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).
Apply f_act_notin_tau.

Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) (cons n (cons n0 l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Inversion_clear H9.
Inversion_clear H12.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H12.
Inversion_clear H14.
Elim (H12 (Out n n0)); Intros.
Cut (ftrans (p x) (Out n n0) (p2 x)); [Intro | Apply H6; Auto].
Elim (H14 (p2 x) H17); Intros.
Inversion_clear H18.
Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19.
Rewrite H22 in H20.
Exists (nu x1); Split.
Apply fRES with empty; Intros; Auto.
(Change (ftrans (q y) ([_:name](Out n n0) y) (x1 y)); Apply FTR_L3 with x; Auto).
Clear H24.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
Inversion_clear H25.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).

Apply f_act_notin_Out; Auto.

Inversion H0.
Generalize H2.
Generalize H0.
Elim a.
Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H10.
Inversion_clear H12.
Elim (H10 tau); Intros.
Cut (ftrans (q x) tau (p2 x)); [Intro | Apply H6; Auto].
Elim (H14 (p2 x) H15); Intros.
Inversion_clear H16.
Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists (nu x1); Split.
Apply fRES with empty; Intros; Auto.
(Change (ftrans (p y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).
Apply f_act_notin_tau.

Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) (cons n (cons n0 l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Inversion_clear H9.
Inversion_clear H12.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H12.
Inversion_clear H14.
Elim (H12 (Out n n0)); Intros.
Cut (ftrans (q x) (Out n n0) (p2 x)); [Intro | Apply H6; Auto].
Elim (H16 (p2 x) H17); Intros.
Inversion_clear H18.
Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19.
Rewrite H22 in H20.
Exists (nu x1); Split.
Apply fRES with empty; Intros; Auto.
Clear H24.
(Change (ftrans (p y) ([_:name](Out n n0) y) (x1 y)); Apply FTR_L3 with x; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
Inversion_clear H25.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).
Apply f_act_notin_Out; Auto.

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H11 x); Intros.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H12 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (q y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
Intro.
(Apply NU_S; Intros).
Elim (LEM_name x0 x3); Intros.
Rewrite <- H22; Auto.
Elim (LEM_name x0 y); Intros.
Rewrite <- H23; Auto.
Rewrite <- H23 in H16.
Rewrite <- H23 in H21.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x3 empty))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Clear H29.
Change (StBisim ([_:name](p2 _ x0) x3) ([_:name](x2 _ x0) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x4 empty))); Intros.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H30.
Inversion_clear H32.
Clear H33.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H28.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H28.
Auto.
Apply notin_nu; Intros.
Inversion_clear H31.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H31.
Auto.
Change (StBisim ([_:name](p2 _ x5) x4) ([_:name](x2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H19.
Auto.
Change (StBisim ([_:name](p2 _ y) x3) ([_:name](x2 _ y) x3)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply b_act_notin_In; Auto.

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H11 x); Intros.
Cut (btrans (q x0) (In x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H14 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (p y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
Intro.
(Apply NU_S; Intros).
Elim (LEM_name x0 x3); Intros.
Rewrite <- H22; Auto.
Elim (LEM_name x0 y); Intros.
Rewrite <- H23; Auto.
Rewrite <- H23 in H16.
Rewrite <- H23 in H21.

Elim (unsat (par (nu [z:name](nu (x2 z))) (nu [z:name](nu (p2 z)))) (cons x0 (cons x3 l))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Change (StBisim ([_:name](x2 _ x0) x3) ([_:name](p2 _ x0) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Elim (unsat (par (nu [z:name](nu (x2 z))) (nu [z:name](nu (p2 z)))) (cons x0 (cons x4 l))); Intros.
Inversion_clear H28.
Inversion_clear H30.
Inversion_clear H31.
Inversion_clear H33.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H28.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H28.
Auto.
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H32.
Auto.
Change (StBisim ([_:name](x2 _ x5) x4) ([_:name](p2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H9.
Auto.
Change (StBisim ([_:name](x2 _ y) x3) ([_:name](p2 _ y) x3)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply b_act_notin_In; Auto.

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H13 x); Intros.
Cut (btrans (p x0) (bOut x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H12 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (q y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
Intros.
(Apply NU_S; Intros).
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x3 (cons y l)))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Inversion_clear H29.
Change (StBisim ([_:name](p2 _ y) x3) ([_:name](x2 _ y) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x4 (cons y l)))); Intros.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Inversion_clear H34.
Inversion_clear H35.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H29.
Auto.
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H33.
Auto.
Change (StBisim ([_:name](p2 _ x5) x4) ([_:name](x2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply H18.
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x0))); [Intro | Auto].
Inversion_clear H29.
Auto.
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x0))); [Intro | Auto].
Inversion_clear H33.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H16.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H16.
Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H21.
Auto.
Apply b_act_notin_bOut; Assumption.

Elim (unsat (par (nu p) (par (nu q) (nu p1))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x1) (q x1)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Elim (H7 (Out x x1)); Intros.
Cut (ftrans (p x1) (Out x x1) (p1 x1)); [Intro | Apply H3; Auto].
Elim (H11 (p1 x1) H14); Intros.
Inversion_clear H15.
Elim (proc_exp x2 x1); Intros.
Inversion_clear H15.
Rewrite H19 in H16.
Rewrite H19 in H17.
Exists x3; Split.
Apply OPEN with empty; Intros; Auto.
Clear H22.
(Change (ftrans (q y) ([_:name](Out x _) y) (x3 y)); Apply FTR_L3 with x1; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Assumption.
Assumption.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Auto.
Assumption.
Intros.
(Apply L6_Light with x1; Auto).

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H13 x); Intros.
Cut (btrans (q x0) (bOut x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H14 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (p y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
Intros.
(Apply NU_S; Intros).
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x3 (cons y l)))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Inversion_clear H29.
Change (StBisim ([_:name](x2 _ y) x3) ([_:name](p2 _ y) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x4 (cons y l)))); Intros.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Inversion_clear H34.
Inversion_clear H35.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H33.
Auto.
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H29.
Auto.
Change (StBisim ([_:name](x2 _ x5) x4) ([_:name](p2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply H18.
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x0))); [Intro | Auto].
Inversion_clear H33.
Auto.
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x0))); [Intro | Auto].
Inversion_clear H29.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H16.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H16.
Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H21.
Auto.
Apply b_act_notin_bOut; Assumption.

Elim (unsat (par (nu p) (par (nu q) (nu q1))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x1) (q x1)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Elim (H7 (Out x x1)); Intros.
Cut (ftrans (q x1) (Out x x1) (q1 x1)); [Intro | Apply H3; Auto].
Elim (H13 (q1 x1) H14); Intros.
Inversion_clear H15.
Elim (proc_exp x2 x1); Intros.
Inversion_clear H15.
Rewrite H19 in H16.
Rewrite H19 in H17.
Exists x3; Split.
Apply OPEN with empty; Intros; Auto.
Clear H22.
(Change (ftrans (p y) ([_:name](Out x _) y) (x3 y)); Apply FTR_L3 with x1; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Assumption.
Assumption.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Auto.
Assumption.
Intros.
(Apply L6_Light with x1; Auto).

Qed.

Transparent NU_S.

Lemma PAR_S: (p,q,r:proc)(StBisim p q)->(StBisim (par p r) (par q r)).

Proof. (* This lemma depends on Lemma NU_S: it is a coinductive proof. *)

(Cofix; Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Inversion H.
Inversion_clear H6.
(Elim (H9 a); Intros).
(Elim (H6 p3 H5); Intros).
(Exists  (par x r); Split; [ Apply fPAR1; Tauto | Apply PAR_S; Tauto ]).

(Exists  (par q p3); Split;
 [ Apply fPAR2; Assumption | Apply PAR_S; Assumption ]).

Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x); Intros).
(Elim (H11 q1 H3); Intros).
Inversion_clear H14.
(Exists  (par (x0 y) q2); Split;
 [ Apply COM1 with x; Assumption | Apply PAR_S; Auto ]).

Inversion H.
Inversion_clear H7.
(Elim (H10 (Out x y)); Intros).
(Elim (H7 q1 H3); Intros).
Inversion_clear H13.
(Exists  (par x0 (q2 y)); Split;
 [ Apply COM2 with x; Assumption | Apply PAR_S; Auto ]).

Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x); Intros).
(Elim (H11 q1 H3); Intros).
Inversion_clear H14.
(Exists  (nu [z:name](par (x0 z) (q2 z))); Split;
 [ Apply CLOSE1 with x; Assumption | Idtac ]).
Apply NU_S; Intros.
Apply PAR_S; Auto.

Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H12 x); Intros).
(Elim (H11 q1 H3); Intros).
Inversion_clear H14.
(Exists  (nu [z:name](par (x0 z) (q2 z))); Split;
 [ Apply CLOSE2 with x; Assumption | Idtac ]).
Apply NU_S; Intros.
Apply PAR_S; Auto.
Apply H16.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin x1 (par (q1 z) (q2 z))); [Intro | Auto].
Inversion_clear H14.
Assumption.
Apply notin_nu; Intros.
Inversion_clear H17.
Cut (notin x1 (par (x0 z) (q2 z))); [Intro | Auto].
Inversion_clear H17.
Assumption.

Inversion H0.
Inversion H.
Inversion_clear H6.
(Elim (H9 a); Intros).
(Elim (H11 p0 H5); Intros).
(Exists  (par x r); Split; [ Apply fPAR1; Tauto | Apply PAR_S; Tauto ]).

(Exists  (par p p0); Split;
 [ Apply fPAR2; Assumption | Apply PAR_S; Assumption ]).

Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x); Intros).
(Elim (H13 q0 H3); Intros).
Inversion_clear H14.
(Exists  (par (x0 y) q2); Split;
 [ Apply COM1 with x; Assumption | Apply PAR_S; Auto ]).

Inversion H.
Inversion_clear H7.
(Elim (H10 (Out x y)); Intros).
(Elim (H12 q0 H3); Intros).
Inversion_clear H13.
(Exists  (par x0 (q2 y)); Split;
 [ Apply COM2 with x; Assumption | Apply PAR_S; Auto ]).

Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x); Intros).
(Elim (H13 q0 H3); Intros).
Inversion_clear H14.
(Exists  (nu [z:name](par (x0 z) (q2 z))); Split;
 [ Apply CLOSE1 with x; Assumption | Idtac ]).
Apply NU_S; Intros.
Apply PAR_S; Auto.

Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H12 x); Intros).
(Elim (H13 q0 H3); Intros).
Inversion_clear H14.
(Exists  (nu [z:name](par (x0 z) (q2 z))); Split;
 [ Apply CLOSE2 with x; Assumption | Idtac ]).
Apply NU_S; Intros.
Apply PAR_S; Auto.
Apply H16.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin x1 (par (x0 z) (q2 z))); [Intro | Auto].
Inversion_clear H14.
Assumption.
Apply notin_nu; Intros.
Inversion_clear H17.
Cut (notin x1 (par (q0 z) (q2 z))); [Intro | Auto].
Inversion_clear H17.
Assumption.

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H6 x); Intros).
(Elim (H10 p3 H5); Intros).
Inversion_clear H13.
Exists [_:name](par (x0 _) r); Split.
Apply bPAR1; Assumption.
Intro; Apply PAR_S; Auto.
Exists [_:name](par q (p3 _)); Split.
Apply bPAR2; Assumption.
Intro; Apply PAR_S; Auto.

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H6 x); Intros).
(Elim (H12 p0 H5); Intros).
Inversion_clear H13.
Exists [_:name](par (x0 _) r); Split.
Apply bPAR1; Assumption.
Intro; Apply PAR_S; Auto.
Exists [_:name](par p (p0 _)); Split.
Apply bPAR2; Assumption.
Intro; Apply PAR_S; Auto.

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H11 x); Intros).
(Elim (H10 p3 H5); Intros).
Inversion_clear H13.
Exists [_:name](par (x0 _) r); Split.
Apply bPAR1; Assumption.
Intros; Apply PAR_S.
Apply H15.
Apply notin_nu; Intros.
Inversion_clear H13.
Cut (notin y (par (p3 z) r)); [Intro | Auto].
Inversion_clear H13.
Assumption.
Apply notin_nu; Intros.
Inversion_clear H16.
Cut (notin y (par (x0 z) r)); [Intro | Auto].
Inversion_clear H16.
Assumption.
Exists [_:name](par q (p3 _)); Split.
Apply bPAR2; Assumption.
Intros; Apply PAR_S.
Assumption.

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H11 x); Intros).
(Elim (H12 p0 H5); Intros).
Inversion_clear H13.
Exists [_:name](par (x0 _) r); Split.
Apply bPAR1; Assumption.
Intros; Apply PAR_S.
Apply H15.
Apply notin_nu; Intros.
Inversion_clear H13.
Cut (notin y (par (x0 z) r)); [Intro | Auto].
Inversion_clear H13.
Assumption.
Apply notin_nu; Intros.
Inversion_clear H16.
Cut (notin y (par (p0 z) r)); [Intro | Auto].
Inversion_clear H16.
Assumption.
Exists [_:name](par p (p0 _)); Split.
Apply bPAR2; Assumption.
Intros; Apply PAR_S.
Assumption.

Qed.

Transparent PAR_S.

Lemma MATCH_S: (p,q:proc)(StBisim p q)->(x,y:name)(StBisim (match x y p) (match x y q)).

Proof. (* This lemma is proved from scratch. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Inversion H.
Inversion_clear H7.
(Elim (H10 a); Intros).
(Elim (H7 p1 H6); Intros).
(Exists  x1; Split; [ Apply fMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
(Elim (H10 a); Intros).
(Elim (H12 q1 H6); Intros).
(Exists  x1; Split; [ Apply fMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x0); Intros).
(Elim (H11 p1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x0); Intros).
(Elim (H13 q1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H12 x0); Intros).
(Elim (H11 p1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intros; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H12 x0); Intros).
(Elim (H13 q1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intros; Auto ]).

Qed.

Transparent MATCH_S.

Lemma MISMATCH_S: (p,q:proc)(StBisim p q)->(x,y:name)(StBisim (mismatch x y p) (mismatch x y q)).

Proof. (* This lemma is proved from scratch. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Inversion H.
Inversion_clear H8.
(Elim (H11 a); Intros).
(Elim (H8 p1 H7); Intros).
(Exists  x1; Split; [ Apply fMISMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
(Elim (H11 a); Intros).
(Elim (H13 q1 H7); Intros).
(Exists  x1; Split; [ Apply fMISMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H8 x0); Intros).
(Elim (H12 p1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H8 x0); Intros).
(Elim (H14 q1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H13 x0); Intros).
(Elim (H12 p1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intros; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H13 x0); Intros).
(Elim (H14 q1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intros; Auto ]).

Qed.

Transparent MISMATCH_S.

Lemma IN_S: (p',q':name->proc)(x,y:name)
            (notin y (nu p'))->(notin y (nu q'))->
            ((z:name)
             ((isin z (nu p'))\/(isin z (nu q'))\/z=y)->
             (StBisim (p' z) (q' z)))->
            (StBisim (in_pref x p') (in_pref x q')).

Proof. (* This lemma depends on Lemma L6_Light. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H2.

Inversion_clear H2.

Inversion H2.
Rewrite <- H3 in H2; Rewrite <- H5 in H2; Rewrite <- H6 in H2.
Exists q'; Split.
Apply IN.
Rewrite <- H3; Intro.
Elim (LEM_OC (par (nu p') (par (nu q') (out_pref y y nil))) y0); Intro.
Inversion_clear H7; [Apply H1; Left; Assumption | Inversion_clear H8; [Apply H1; Right; Left; Assumption | Inversion H7; [Inversion_clear H9 | Apply H1; Right; Right; Trivial | Apply H1; Right; Right; Trivial]]].
Inversion_clear H7.
Inversion_clear H9.
Inversion_clear H10.
Clear H11 H12.
Apply L6_Light with y; Auto.

Inversion H2.
Rewrite <- H3 in H2; Rewrite <- H5 in H2; Rewrite <- H6 in H2.
Exists p'; Split.
Apply IN.
Rewrite <- H3; Intro.
Elim (LEM_OC (par (nu p') (par (nu q') (out_pref y y nil))) y0); Intro.
Inversion_clear H7; [Apply H1; Left; Assumption | Inversion_clear H8; [Apply H1; Right; Left; Assumption | Inversion H7; [Inversion_clear H9 | Apply H1; Right; Right; Trivial | Apply H1; Right; Right; Trivial]]].
Inversion_clear H7.
Inversion_clear H9.
Inversion_clear H10.
Clear H11 H12.
Apply L6_Light with y; Auto.

Inversion_clear H2.

Inversion_clear H2.

Qed.

Transparent IN_S.

Lemma OUT_S: (p,q:proc)(StBisim p q)->(x,y:name)(StBisim (out_pref x y p) (out_pref x y q)).

Proof. (* This lemma is proved from scratch. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
(Exists  q; Split; [ Apply OUT | Rewrite <- H5; Assumption ]).

Inversion H0.
(Exists  p; Split; [ Apply OUT | Rewrite <- H5; Assumption ]).

Inversion_clear H0.

Inversion_clear H0.

Inversion_clear H0.

Inversion_clear H0.

Qed.

Transparent OUT_S.

End structural_congruence1.

Section monoidal_sum.

Lemma ID_SUM: (p:proc)(StBisim (sum p nil) p).

Proof. (* This lemma depends on Lemma REF. *)

Intro; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Assumption | Apply REF].
Inversion_clear H0.
Exists q1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intro; Apply REF].
Inversion_clear H0.
Exists q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intros; Apply REF].
Inversion_clear H0.
Exists q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Qed.

Lemma IDEM_SUM: (p:proc)(StBisim (sum p p) p).

Proof. (* This lemma depends on Lemma REF. *)

Intro; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Assumption | Apply REF].
Exists  p1; Split; [Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intro; Apply REF].
Exists  p1; Split; [Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intros; Apply REF].
Exists  p1; Split; [Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Qed.

Lemma COMM_SUM: (p,q:proc)(StBisim (sum p q) (sum q p)).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Apply fSUM2; Assumption | Apply REF].
Exists  p1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  q1; Split; [Apply fSUM2; Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM2; Assumption | Intro; Apply REF].
Exists  p1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  q1; Split; [Apply bSUM2; Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM2; Assumption | Intros; Apply REF].
Exists  p1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Inversion_clear H.
Exists  q1; Split; [Apply bSUM2; Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Qed.

Lemma ASS_SUM: (p,q,r:proc)(StBisim (sum p (sum q r)) (sum (sum p q) r)).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Apply fSUM1; Apply fSUM1; Assumption | Apply REF].
Inversion_clear H0.
Exists  p1; Split; [Apply fSUM1; Apply fSUM2; Assumption | Apply REF].
Exists  p1; Split; [Apply fSUM2; Assumption | Apply REF].

Inversion_clear H.
Inversion_clear H0.
Exists  q1; Split; [Apply fSUM1; Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM2; Apply fSUM1; Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM2; Apply fSUM2; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM1; Apply bSUM1; Assumption | Intro; Apply REF].
Inversion_clear H0.
Exists  p1; Split; [Apply bSUM1; Apply bSUM2; Assumption | Intro; Apply REF].
Exists  p1; Split; [Apply bSUM2; Assumption | Intro; Apply REF].

Inversion_clear H.
Inversion_clear H0.
Exists  q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM1; Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM2; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM1; Apply bSUM1; Assumption | Intros; Apply REF].
Inversion_clear H0.
Exists  p1; Split; [Apply bSUM1; Apply bSUM2; Assumption | Intros; Apply REF].
Exists  p1; Split; [Apply bSUM2; Assumption | Intros; Apply REF].

Inversion_clear H.
Inversion_clear H0.
Exists  q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM1; Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM2; Assumption | Intros; Apply REF].

Qed.

End monoidal_sum.

Section bang_unfolding.

Lemma BANG_UNF : (p:proc)(StBisim (bang p) (par p (bang p))).

Proof. (* This lemma is proved from scratch. *)

Intro; Apply sb; Do 3 Try (Split; Intros).
Exists p1; Split; [Inversion_clear H; Assumption | Apply REF].
Exists q1; Split; [Apply fBANG; Assumption | Apply REF].
Exists p1; Split; [Inversion_clear H; Assumption | Intro; Apply REF].
Exists q1; Split; [Apply bBANG; Assumption | Intro; Apply REF].
Exists p1; Split; [Inversion_clear H; Assumption | Intros; Apply REF].
Exists q1; Split; [Apply bBANG; Assumption | Intros; Apply REF].

Qed.

End bang_unfolding.

Section restriction_laws1.

Lemma NU_NIL: (StBisim (nu [x:name]nil) nil).

Proof. (* This lemma is proved from scratch. *)

Apply sb; Do 3 Try (Split; Intros).

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans nil tau (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans nil (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Apply f_act_notin_Out; Auto.

Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans nil (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.

Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans nil (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [_:name]nil) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans nil (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.

Inversion_clear H.

Qed.

Lemma NU_NIL1: (p:proc)(x:name)(StBisim (nu [y:name](out_pref y x p)) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).
Generalize H; Clear H; Elim a; Intros.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (out_pref x0 x p) tau (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (out_pref x0 x p) (Out n n0) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H4.
Absurd x0=n; Auto.
Apply f_act_notin_Out; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x1 x p) (In x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x1 x p) (bOut x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu p1)) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (out_pref x1 x p) (Out x0 x1) (p1 x1)); [Intro | Apply H0; Auto].
Inversion H2.
Absurd x1=x0; Auto.
Inversion_clear H.

Qed.

Lemma NU_NIL2: (p:name->proc)(StBisim (nu [y:name](in_pref y p)) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).
Generalize H; Clear H; Elim a; Intros.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (in_pref x p) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (in_pref x p) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.
Apply f_act_notin_Out; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x0 p) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Absurd x0=x; Auto.
Apply b_act_notin_In; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x0 p) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (in_pref x0 p) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Inversion_clear H.

Qed.

Lemma eta_slb: (p:name->proc)(StBisim (nu p) (nu [x:name](p x))).

Proof. (* This lemma depends on Lemmata FTR_L3, BTR_L3, REF:
        *  it is a coinductive proof.
        *)

Cofix; Intros; Apply sb; Do 3 Try (Split; Intros).

Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu p) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (p x) tau (p2 x)); [Intro | Apply H0; Auto].
Exists (nu [x:name](p2 x)); Split.
Apply fRES with (cons x empty); Intros.
Inversion_clear H6.
Clear H7 H9.
Change (ftrans (p y) ([_:name]tau y) (p2 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H4; Apply notin_nu; Intros; Auto.
Inversion_clear H5; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply eta_slb.
Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu p) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (p x) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Exists (nu [x:name](p2 x)); Split.
Apply fRES with (cons x empty); Intros.
Inversion_clear H8.
Inversion_clear H9.
Clear H11.
Change (ftrans (p y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H6; Apply notin_nu; Intros; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply eta_slb.
Apply f_act_notin_Out; Auto.

(* The way back: symmetric case *)

Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu [x:name](p x)) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (p x) tau (p2 x)); [Intro | Apply H0; Auto].
Exists (nu p2); Split.
Apply fRES with (cons x empty); Intros.
Inversion_clear H6.
Clear H7 H9.
Change (ftrans (p y) ([_:name]tau y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply REF.
Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [x:name](p x)) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (p x) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Exists (nu p2); Split.
Apply fRES with (cons x empty); Intros.
Inversion_clear H8.
Inversion_clear H9.
Clear H11.
Change (ftrans (p y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply REF.
Apply f_act_notin_Out; Auto.

(* INPUT actions case *)

Inversion_clear H.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Exists [_:name](nu [x0:name](p2 x0 _)); Split.
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H7; Inversion_clear H8.
Clear H10.
Change (btrans (p y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Inversion_clear H5; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Apply REF.
Apply b_act_notin_In; Auto.

(* The way back: symmetric case *)

Inversion_clear H.
Elim (unsat (par (nu [x0:name](p x0)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Exists [_:name](nu [x0:name](p2 x0 _)); Split.
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H7; Inversion_clear H8.
Clear H10.
Change (btrans (p y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Apply REF.
Apply b_act_notin_In; Auto.

(* BOUND OUTPUT actions case *)

Inversion_clear H.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (p x0) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Exists [_:name](nu [x0:name](p2 x0 _)); Split.
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H7; Inversion_clear H8.
Clear H10.
Change (btrans (p y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Inversion_clear H5; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu p) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (p x0) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Exists p1; Split.
Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H8.
Clear H10.
Change (ftrans (p y) (Out x y) (p1 y)); Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H5; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.

(* The way back: symmetric case *)

Inversion_clear H.
Elim (unsat (par (nu [x0:name](p x0)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (p x0) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Exists [_:name](nu [x0:name](p2 x0 _)); Split.
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H7; Inversion_clear H8.
Clear H10.
Change (btrans (p y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [x0:name](p x0)) (nu q1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (p x0) (Out x x0) (q1 x0)); [Intro | Apply H0; Auto].
Exists q1; Split.
Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H8.
Clear H10.
Change (ftrans (p y) (Out x y) (q1 y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.

Qed.

Lemma NU_COMM: (p:name->name->proc)(StBisim (nu [y:name](nu [z:name](p y z))) (nu [z:name](nu [y:name](p y z)))).

Proof. (* This lemma depends on Lemmata FTR_L3, BTR_L3, SYM, TRANS,
        * NU_S, eta_slb: it is a coinductive proof.
        *)

Cofix; Intro; Apply sb; Do 3 Try (Split; Intros).
Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (nu [z:name](p x z)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.
Elim (unsat (par (nu [z:name](p x z)) (nu p3)) l0); Intros.
Inversion_clear H8.
Inversion_clear H9.
Cut (ftrans (p x x0) tau (p3 x0)); [Intro | Apply H7; Auto].
Elim (ho_proc_exp p3 x); Intros.
Inversion_clear H12.
Exists (nu [z:name](nu [y:name](x1 y z))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H16.
Inversion_clear H19.
Clear H17 H20.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20.
Inversion_clear H23.
Clear H21 H24.
Change (ftrans ([_:name](p _ y) y0) ([_:name]tau y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H13.
Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Change (ftrans (p x y) ([_:name]tau y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H8; Apply notin_nu; Intros; Auto.
Rewrite <- H14; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite <- H14; Assumption.
Inversion_clear H12.
Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H15.
Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x1 y0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite H14 in H5.
Cut ([_:name](nu (x1 _)))=p2; [Intro | Apply proc_ext with x; Auto].
Rewrite <- H12.
Apply TRANS with (nu [_:name](nu [x:name](x1 _ x))); [Apply NU_S; Intros; Apply eta_slb | Apply NU_COMM].
Apply f_act_notin_tau.
Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (nu [z:name](p x z)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.
Elim (unsat (par (nu [z:name](p x z)) (nu p3)) (cons n (cons n0 l0))); Intros.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H14.
Cut (ftrans (p x x0) (Out n n0) (p3 x0)); [Intro | Apply H9; Auto].
Elim (ho_proc_exp p3 x); Intros.
Inversion_clear H16.
Exists (nu [z:name](nu [y:name](x1 y z))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20.
Inversion_clear H23.
Inversion_clear H21.
Clear H24.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H26.
Inversion_clear H29.
Inversion_clear H27.
Clear H30.
Change (ftrans ([_:name](p _ y) y0) ([_:name](Out n n0) y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H30; Auto.
Inversion_clear H17.
Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H30; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans (p x y) ([_:name](Out n n0) y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H10; Apply notin_nu; Intros; Auto.
Rewrite <- H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H18; Assumption.
Inversion_clear H16.
Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H30; Auto.
Inversion_clear H19.
Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x1 y0 z))); [Intro | Auto].
Inversion_clear H30; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H18 in H7.
Cut ([_:name](nu (x1 _)))=p2; [Intro | Apply proc_ext with x; Auto].
Rewrite <- H16.
Apply TRANS with (nu [_:name](nu [x:name](x1 _ x))); [Apply NU_S; Intros; Apply eta_slb | Apply NU_COMM].
Apply f_act_notin_Out; Auto.
Apply f_act_notin_Out; Auto.

(* The way back: symmetric case *)

Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (nu [y0:name](p y0 x)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.
Elim (unsat (par (nu [y0:name](p y0 x)) (nu p0)) l0); Intros.
Inversion_clear H8.
Inversion_clear H9.
Cut (ftrans (p x0 x) tau (p0 x0)); [Intro | Apply H7; Auto].
Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H12.
Exists (nu [y:name](nu [z:name](x1 z y))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H16.
Inversion_clear H19.
Clear H17 H20.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20.
Inversion_clear H23.
Clear H21 H24.
Change (ftrans (p y y0) ([_:name]tau y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H13.
Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Change (ftrans ([_:name](p _ x) y) ([_:name]tau y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Rewrite <- H14; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite <- H14; Assumption.
Inversion_clear H12.
Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H15.
Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x1 z0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H17; Apply notin_nu; Intros ; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Cut ([_:name](nu (x1 _)))=p2; [Intro | Apply proc_ext with x; Auto].
Rewrite <- H12.
Apply SYM; Apply TRANS with (nu [_:name](nu [x:name](x1 _ x))); [Apply NU_S; Intros; Apply eta_slb | Apply NU_COMM].
Rewrite <- H14; Assumption.
Apply f_act_notin_tau.
Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (nu [y0:name](p y0 x)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.
Elim (unsat (par (nu [y0:name](p y0 x)) (nu p0)) (cons n (cons n0 l0))); Intros.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H14.
Cut (ftrans (p x0 x) (Out n n0) (p0 x0)); [Intro | Apply H9; Auto].
Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H16.
Exists (nu [y:name](nu [z:name](x1 z y))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20.
Inversion_clear H23.
Inversion_clear H21.
Clear H24.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H26.
Inversion_clear H29.
Inversion_clear H27.
Clear H30.
Change (ftrans (p y y0) ([_:name](Out n n0) y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x (nu [z0:name](p z0 z))); [Intro | Auto].
Inversion_clear H30; Auto.
Inversion_clear H17.
Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H30; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans ([_:name](p _ x) y) ([_:name](Out n n0) y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Rewrite <- H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H18; Assumption.
Inversion_clear H16.
Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H30; Auto.
Inversion_clear H19.
Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x1 z0 z))); [Intro | Auto].
Inversion_clear H30; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H18 in H7.
Cut ([_:name](nu (x1 _)))=p2; [Intro | Apply proc_ext with x; Auto].
Rewrite <- H16.
Apply SYM; Apply TRANS with (nu [_:name](nu [x:name](x1 _ x))); [Apply NU_S; Intros; Apply eta_slb | Apply NU_COMM].
Apply f_act_notin_Out; Auto.
Apply f_act_notin_Out; Auto.

(* INPUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (nu [z:name](p x0 z)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [z:name](p x0 z)) (nu [z:name](nu (p3 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (btrans (p x0 x1) (In x) (p3 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p3 x0); Intros.
Inversion_clear H14.
Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [z:name](nu [y:name](p y z))) (In x)
     [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); 
Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H21.
Clear H22.
Change  (btrans (nu [y0:name](p y0 y)) (In x)
     [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); 
Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H26.
Clear H27.
Change (btrans ([_:name](p _ y) y0) ([_:name](In x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (btrans (p x0 y) ([_:name](In x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14.
Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H16 in H6.
Cut ([_:name][y:name](nu [z:name](x2 _ z y)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite <- H14; Intro.
Change (StBisim (nu [z:name](nu [z0:name]([u:name][v:name](x2 u v y) z z0)))
     (nu [z:name](nu [y0:name]([u:name][v:name](x2 u v y) y0 z))));
Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_In; Auto.
Apply b_act_notin_In; Auto.

(* The way back: symmetric case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (nu [y0:name](p y0 x0)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [y0:name](p y0 x0)) (nu [z:name](nu (p0 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (btrans (p x1 x0) (In x) (p0 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p0 x0); Intros.
Inversion_clear H14.
Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [z:name](nu [y:name](p z y))) (In x)
     [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); 
Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H21.
Clear H22.
Change  (btrans (nu [y0:name](p y y0)) (In x)
     [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); 
Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H26.
Clear H27.
Change (btrans (p y y0) ([_:name](In x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (btrans ([_:name](p _ x0) y) ([_:name](In x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14.
Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p z y0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H16 in H6.
Cut ([_:name][y:name](nu [z:name](x2 _ z y)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite <- H14; Intro.
Apply SYM; Change (StBisim (nu [z:name](nu [z0:name]([u:name][v:name](x2 u v y) z z0))) (nu [z:name](nu [y0:name]([u:name][v:name](x2 u v y) y0 z))));
Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_In; Auto.
Apply b_act_notin_In; Auto.

(* BOUND OUTPUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (nu [z:name](p x0 z)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [z:name](p x0 z)) (nu [z:name](nu (p3 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (btrans (p x0 x1) (bOut x) (p3 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p3 x0); Intros.
Inversion_clear H14.
Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [z:name](nu [y:name](p y z))) (bOut x)
     [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); 
Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H21.
Clear H22.
Change (btrans (nu [y0:name](p y0 y)) (bOut x)
     [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); 
Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H26.
Clear H27.
Change (btrans ([_:name](p _ y) y0) ([_:name](bOut x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Change (btrans (p x0 y) ([_:name](bOut x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14.
Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H16 in H6.
Cut ([_:name][y:name](nu [z:name](x2 _ z y)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite <- H14; Intros.
Change (StBisim (nu [z:name](nu [z0:name]([u:name][v:name](x2 u v y) z z0))) (nu [z:name](nu [y0:name]([u:name][v:name](x2 u v y) y0 z))));
Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [z:name](p x0 z)) (nu (p2 x0))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (ftrans (p x0 x2) (Out x x2) (p2 x0 x2)); [Intro | Apply H7; Auto].
Exists [_:name](nu [z:name](p2 z _)); Split.
Apply OPEN with (cons x (cons x0 empty)); Intros.
Inversion_clear H17.
Inversion_clear H19.
Clear H20.
Apply fRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H24.
Clear H26.
Change (ftrans ([_:name](p _ y) y0) ([_:name](Out x y) y0) ([_:name](p2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H3.
Apply notin_nu; Intros.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans (p x0 y) (Out x y) (p2 x0 y)); Apply FTR_L3 with x2; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H14.
Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (nu [z:name](p x0 z)) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [z:name](p x0 z)) (nu p2)) (cons x (cons x0 l0))); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H13.
Cut (ftrans (p x0 x1) (Out x x0) (p2 x1)); [Intro | Apply H8; Auto].
Elim (ho_proc_exp p2 x0); Intros.
Inversion_clear H15.
Exists [y:name](nu [z:name](x2 y z)); Split.
Change (btrans (nu [z:name](nu [y:name](p y z))) (bOut x)
        [y:name](nu [z:name]([u:name][_:name](x2 _ u) z y)));
Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H22.
Clear H23.
Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H25.
Clear H27.
Change (ftrans ([_:name](p _ y) y0) (Out x y0) ([_:name](x2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H16; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans (p x0 y) ([_:name](Out x x0) y) (x2 x0 y)); Apply FTR_L3 with x1; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Rewrite <- H17; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H17; Assumption.
Inversion_clear H15.
Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [_:name](x2 _ z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H17 in H6.
Cut ([_:name](nu (x2 _)))=p1; [Intro | Apply proc_ext with x0; Auto].
Rewrite <- H15; Intros; Apply eta_slb.
Apply f_act_notin_Out; Auto.

(* The way back: symmetric case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (nu [y0:name](p y0 x0)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [y0:name](p y0 x0)) (nu [z:name](nu (p0 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (btrans (p x1 x0) (bOut x) (p0 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p0 x0); Intros.
Inversion_clear H14.
Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [y:name](nu [z:name](p y z))) (bOut x)
     [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); 
Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18.
Inversion_clear H19.
Inversion_clear H21.
Clear H22.
Change (btrans (nu [z:name](p y z)) (bOut x)
     [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); 
Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H26.
Clear H27.
Change (btrans (p y y0) ([_:name](bOut x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15.
Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Change (btrans ([_:name](p _ x0) y) ([_:name](bOut x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14.
Apply notin_nu; Intros; Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H16 in H6.
Cut ([_:name][y:name](nu [z:name](x2 _ z y)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite <- H14; Intros.
Change (StBisim (nu [z:name](nu [y0:name]([u:name][v:name](x2 v u y) z y0))) (nu [y0:name](nu [z:name]([u:name][v:name](x2 v u y) z y0))));
Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](p y0 x0)) (nu (p2 x0))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Cut (ftrans (p x2 x0) (Out x x2) (p2 x0 x2)); [Intro | Apply H7; Auto].
Exists [_:name](nu [z:name](p2 z _)); Split.
Apply OPEN with (cons x (cons x0 empty)); Intros.
Inversion_clear H17.
Inversion_clear H19.
Clear H20.
Apply fRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H21.
Inversion_clear H22.
Inversion_clear H24.
Clear H26.
Change (ftrans (p y y0) ([_:name](Out x y) y0) ([_:name](p2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H3.
Apply notin_nu; Intros.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans ([_:name](p _ x0) y) (Out x y) (p2 x0 y)); Apply FTR_L3 with x2; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H14.
Apply notin_nu; Intros; Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H19; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu q1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (nu [y0:name](p y0 x0)) (Out x x0) (q1 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [y0:name](p y0 x0)) (nu p2)) (cons x (cons x0 l0))); Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H13.
Cut (ftrans (p x1 x0) (Out x x0) (p2 x1)); [Intro | Apply H8; Auto].
Elim (ho_proc_exp p2 x0); Intros.
Inversion_clear H15.
Exists [y:name](nu [z:name](x2 y z)); Split.
Change (btrans (nu [y:name](nu [z:name](p y z))) (bOut x)
        [y:name](nu [z:name]([u:name][_:name](x2 _ u) z y)));
Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H19.
Inversion_clear H20.
Inversion_clear H22.
Clear H23.
Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H25.
Clear H27.
Change (ftrans (p y y0) (Out x y0) ([_:name](x2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H.
Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H16; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans ([_:name](p _ x0) y) ([_:name](Out x x0) y) (x2 x0 y)); Apply FTR_L3 with x1; Auto.
Rewrite <- H17; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H17; Assumption.
Inversion_clear H15.
Apply notin_nu; Intros; Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [_:name](x2 _ z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H22; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H17 in H6.
Cut ([_:name](nu (x2 _)))=q1; [Intro | Apply proc_ext with x0; Auto].
Rewrite <- H15; Intros; Apply SYM; Apply eta_slb.
Apply f_act_notin_Out; Auto.

Qed.

Lemma NU_EXTR: (p:name->proc)(q:proc)(StBisim (nu [y:name](par (p y) q)) (par (nu p) q)).

Proof. (* This lemma depends on Lemmata FTR_L1, BTR_L1, FTR_L3, BTR_L3,
        * REF, TRANS, NU_S, NU_COMM: it is a coinductive proof.
        *)

Cofix; Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](par (p y0) q)) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (par (p x) q) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.

Elim (proc_exp p4 x); Intros.
Inversion_clear H9.
Exists (par (nu x0) q); Split.
Apply fPAR1.
Apply fRES with (cons x empty); Intros.
Inversion_clear H13.
Clear  H14 H16.
Change (ftrans (p y) ([_:name]tau y) (x0 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y.
Cut (notin x (par (p y) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau; Auto.
Rewrite <- H11; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau; Auto.
Rewrite H11 in H8.
Cut ([_:name](par (x0 _) q))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <-H9; Apply NU_EXTR.
Apply notin_nu; Intros.
Apply notin_par; [Inversion_clear H10; Auto | Inversion_clear H; Cut (notin x (par (p z) q)); [Intro | Auto]; Inversion_clear H; Assumption].

Exists (par (nu p) p4); Split.
Apply fPAR2; Assumption.
Cut ([_:name](par (p _) p4))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <-H9; Apply NU_EXTR.
Apply notin_nu; Intros; Inversion_clear H; Cut (notin x (par (p z) q)); [Intro | Auto]; Inversion_clear H; Apply notin_par; [Assumption | Elim (FTR_L1 q p4 tau x H12 H7); Intros; Assumption].

Elim (ho_proc_exp q1 x); Intros.
Inversion_clear H10.
Rewrite H12 in H8.
Exists (par (nu [_:name](x1 _ y)) q2); Split.
Change (ftrans (par (nu p) q) tau (par ([z:name](nu [_:name](x1 _ z)) y) q2)); Apply COM1 with x0.
Apply bRES with (cons x empty); Intros.
Inversion_clear H15.
Clear H17.
Change (btrans (p y0) ([_:name](In x0) y0) (x1 y0)); Apply BTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y0.
Cut (notin x (par (p y0) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Inversion_clear H.
Cut (notin x (par (p y0) q)); [Intro | Auto].
Inversion_clear H.
Elim (FTR_L1 q q2 (Out x0 y) x H18 H9); Intros.
Inversion_clear H; Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Assumption.
Assumption.
Rewrite H12 in H6; Cut ([_:name](par (x1 _ y) q2))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <-H10; Change (StBisim (nu [_:name](par ([z:name](x1 z y) _) q2)) (par (nu [_:name](x1 _ y)) q2)); Apply NU_EXTR.
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H14.
Clear H13 H15.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H13.
Elim (FTR_L1 q q2 (Out x0 y) x H15 H9); Intros.
Inversion_clear H13.
Inversion_clear H11; Apply notin_nu; Intros; Apply notin_par; [Cut (notin x (nu (x1 z))); [Intro | Auto]; Inversion_clear H19; Auto | Assumption].

Elim (proc_exp q1 x); Intros.
Inversion_clear H10.
Rewrite H12 in H8.
Elim (LEM_name x y); Intro.

Rewrite <- H10 in H8; Rewrite <- H10 in H6.
Exists (nu [_:name](par (x1 _) (q2 _))); Split.
Apply CLOSE2 with x0.
Apply OPEN with (cons x empty); Intros.
Inversion_clear H16.
Clear H18.
Change (ftrans (p y0) ([_:name](Out x0 _) y0) (x1 y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y0.
Cut (notin x (par (p y0) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Inversion_clear H.
Cut (notin x (par (p y0) q)); [Intro | Auto].
Inversion_clear H.
Elim (BTR_L1 q q2 (In x0) x H19 H9); Intros.
Inversion_clear H.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Assumption.
Rewrite H12 in H6; Cut ([_:name](par (x1 _) (q2 _)))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <- H13; Apply REF.
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H15.
Clear H16.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H15.
Elim (BTR_L1 q q2 (In x0) x H17 H9); Intros.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H11; Auto | Inversion_clear H18; Auto].

Exists (par (nu x1) (q2 y)); Split.
Apply COM2 with x0.
Apply fRES with (cons x empty); Intros.
Inversion_clear H16.
Inversion_clear H15.
Clear H19.
Change (ftrans (p y0) ([_:name](Out x0 y) y0) (x1 y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y0.
Cut (notin x (par (p y0) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Inversion_clear H.
Cut (notin x (par (p y0) q)); [Intro | Auto].
Inversion_clear H.
Elim (BTR_L1 q q2 (In x0) x H20 H9); Intros.
Inversion_clear H.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Assumption.
Rewrite H12 in H6; Cut ([_:name](par (x1 _) (q2 y)))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <- H13; Apply NU_EXTR.
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H15.
Clear H16.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H15.
Elim (BTR_L1 q q2 (In x0) x H17 H9); Intros.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H11; Auto | Inversion_clear H18; Auto].

Elim (ho_proc_exp q1 x); Intros.
Inversion_clear H10.
Exists (nu [z:name](par (nu [_:name](x1 _ z)) (q2 z))); Split.
Change (ftrans (par (nu p) q) tau
     (nu [z:name](par ([i:name](nu [_:name](x1 _ i)) z) (q2 z)))); Apply CLOSE1 with x0.
Apply bRES with (cons x empty); Intros.
Inversion_clear H15.
Clear  H17.
Change (btrans (p y) ([_:name](In x0) y) (x1 y)); Apply BTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y.
Cut (notin x (par (p y) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H18.
Clear H19.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H18.
Elim (BTR_L1 q q2 (bOut x0) x H20 H9); Intros.
Inversion_clear H18.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite <- H12; Assumption.
Inversion_clear H14; Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Assumption.
Rewrite H12 in H6.
Cut ([_:name](nu [z:name](par (x1 _ z) (q2 z))))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <- H10; Apply TRANS with (nu [z:name](nu [_:name](par (x1 _ z) (q2 z))));
[Change (StBisim (nu [_:name](nu [z:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z)))
     (nu [z:name](nu [_:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z))));
Apply NU_COMM |
Apply NU_S; Intros; Change (StBisim (nu [_:name](par ([u:name](x1 u x2) _) (q2 x2)))
     (par (nu [_:name](x1 _ x2)) (q2 x2))); Apply NU_EXTR].
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H14.
Clear H15.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H14.
Elim (BTR_L1 q q2 (bOut x0) x H16 H9); Intros.
Inversion_clear H11.
Apply notin_nu; Intros; Apply notin_nu; Intros; Cut (notin x (nu (x1 z))); [Intro | Auto]; Inversion_clear H17; Inversion_clear H20; Apply notin_par; Auto.

Elim (ho_proc_exp q1 x); Intros.
Inversion_clear H10.
Exists (nu [z:name](par (nu [_:name](x1 _ z)) (q2 z))); Split.
Change (ftrans (par (nu p) q) tau
     (nu [z:name](par ([i:name](nu [_:name](x1 _ i)) z) (q2 z)))); Apply CLOSE2 with x0.
Apply bRES with (cons x empty); Intros.
Inversion_clear H15.
Clear  H17.
Change (btrans (p y) ([_:name](bOut x0) y) (x1 y)); Apply BTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y.
Cut (notin x (par (p y) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H18.
Clear H19.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H18.
Elim (BTR_L1 q q2 (In x0) x H20 H9); Intros.
Inversion_clear H18.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite <- H12; Assumption.
Inversion_clear H14; Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Assumption.
Rewrite H12 in H6.
Cut ([_:name](nu [z:name](par (x1 _ z) (q2 z))))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <- H10; Apply TRANS with (nu [z:name](nu [_:name](par (x1 _ z) (q2 z))));
[Change (StBisim (nu [_:name](nu [z:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z)))
     (nu [z:name](nu [_:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z))));
Apply NU_COMM |
Apply NU_S; Intros; Change (StBisim (nu [_:name](par ([u:name](x1 u x2) _) (q2 x2)))
     (par (nu [_:name](x1 _ x2)) (q2 x2))); Apply NU_EXTR].
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H14.
Clear H15.
Cut (notin x (par (p x2) q)); [Intro | Auto].
Inversion_clear H14.
Elim (BTR_L1 q q2 (In x0) x H16 H9); Intros.
Inversion_clear H11.
Apply notin_nu; Intros; Apply notin_nu; Intros; Cut (notin x (nu (x1 z))); [Intro | Auto]; Inversion_clear H17; Inversion_clear H20; Apply notin_par; Auto.

Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](par (p y0) q)) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (par (p x) q) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.

Elim (proc_exp p4 x); Intros.
Inversion_clear H11.
Exists (par (nu x0) q); Split.
Apply fPAR1.
Apply fRES with (cons x empty); Intros.
Inversion_clear H15; Inversion_clear H16.
Clear H18.
Change (ftrans (p y) ([_:name](Out n n0) y) (x0 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H.
Apply proc_mono with y.
Cut (notin x (par (p y) q)); [Intro | Auto].
Inversion_clear H; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H13; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H13 in H10.
Cut ([_:name](par (x0 _) q))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <- H11; Apply NU_EXTR.
Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H15.
Clear H14 H16.
Cut (notin x (par (p x1) q)); [Intro | Auto].
Inversion_clear H14.
Inversion_clear H12.
Apply notin_nu; Intros; Apply notin_par; Auto.

Exists (par (nu p) p4); Split; [Apply fPAR2; Assumption | Idtac].
Cut ([_:name](par (p _) p4))=p2; [Intro | Apply proc_ext with x; Intros; Auto].
Rewrite <- H11; Apply NU_EXTR.

Inversion_clear H.
Elim (unsat nil (cons x empty)); Intros.
Inversion_clear H.
Inversion_clear H13.
Clear H12 H14.
Cut (notin x (par (p x0) q)); [Intro | Auto].
Inversion_clear H12.
Elim (FTR_L1 q p4 (Out n n0) x H14 H9); Intros.
Apply proc_mono with x0; Apply notin_par; Assumption.
Apply f_act_notin_Out; Auto.

(* The way back: symmetric case *)

Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Inversion_clear H0.
Exists (nu [_:name](par (p2 _) q)); Split.
Apply fRES with l; Intros.
Apply fPAR1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H5; Assumption.
Inversion_clear H1; Apply notin_nu; Intros.
Cut (notin y (par (p2 z) q)); [Intro | Auto].
Inversion_clear H5; Assumption.
Apply NU_EXTR.

Exists (nu [_:name](par (p _) p0)); Split.
Apply fRES with empty; Intros.
Clear H2 H3.
Apply fPAR2; Assumption.
Apply NU_EXTR.

Inversion_clear H0.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x (cons y l))); Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H5.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H; Auto].
Exists (nu [y0:name](par (p2 y0 y) q2)); Split.
Apply fRES with (cons x (cons x0 (cons y empty))); Intros.
Inversion_clear H9.
Inversion_clear H12.
Inversion_clear H13.
Clear  H10 H14.
Apply COM1 with x.
Change (btrans (p y0) ([_:name](In x) y0) (p2 y0)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Inversion_clear H7; Apply notin_nu; Intros.
Cut (notin y0 (par (p z) q)); [Intro | Auto].
Inversion_clear H13; Auto.
Inversion_clear H8; Apply notin_nu; Intros.
Cut (notin y0 (par (p2 z y) q2)); [Intro | Auto].
Apply proc_mono with y; Inversion_clear H13; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Assumption.
Change (StBisim (nu [y0:name](par ([_:name](p2 _ y) y0) q2)) (par (nu [z:name](p2 z y)) q2));
Apply NU_EXTR.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.

Inversion_clear H0.
Elim (unsat (par (nu p) (nu p2)) (cons x (cons y l))); Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H5.
Cut (ftrans (p x0) (Out x y) (p2 x0)); [Intro | Apply H; Auto].
Exists (nu [y0:name](par (p2 y0) (q2 y))); Split.
Apply fRES with (cons x (cons x0 (cons y empty))); Intros.
Inversion_clear H9.
Inversion_clear H12.
Inversion_clear H13.
Clear  H10 H14.
Apply COM2 with x.
Change (ftrans (p y0) ([_:name](Out x y) y0) (p2 y0)); Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H7; Apply notin_nu; Intros.
Cut (notin y0 (par (p z) q)); [Intro | Auto].
Inversion_clear H13; Auto.
Inversion_clear H8; Apply notin_nu; Intros.
Cut (notin y0 (par (p2 z) (q2 y))); [Intro | Auto].
Inversion_clear H13; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Assumption.
Apply NU_EXTR.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.

Inversion_clear H0.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H; Auto].
Exists (nu [y:name](nu [_:name](par (p2 y _) (q2 _)))); Split.
Apply fRES with (cons x (cons x0 l)); Intros.
Inversion_clear H8.
Inversion_clear H11.
Clear H9.
Apply CLOSE1 with x.
Change (btrans (p y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Inversion_clear H6; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H11; Auto.
Inversion_clear H7; Apply notin_nu; Intros.
Cut (notin y (nu [_:name](par (p2 z _) (q2 _)))); [Intro | Auto].
Inversion_clear H11; Apply notin_nu; Intros.
Cut (notin y (par (p2 z z0) (q2 z0))); [Intro | Auto].
Inversion_clear H14; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Assumption.
Apply TRANS with (nu [_:name](nu [y:name](par (p2 y _) (q2 _))));
[Change (StBisim (nu [y:name](nu [_:name]([u:name][v:name](par (p2 u v) (q2 v)) y _)))
                (nu [_:name](nu [y:name]([u:name][v:name](par (p2 u v) (q2 v)) y _))));
 Apply NU_COMM |
 Apply NU_S; Intros;
 Change (StBisim (nu [y:name](par ([_:name](p2 _ x1) y) (q2 x1)))
                 (par (nu [z:name](p2 z x1)) (q2 x1))); Apply NU_EXTR
].
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.

Inversion_clear H0.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Cut (btrans (p x0) (bOut x) (p2 x0)); [Intro | Apply H; Auto].
Exists (nu [y:name](nu [_:name](par (p2 y _) (q2 _)))); Split.
Apply fRES with (cons x (cons x0 l)); Intros.
Inversion_clear H8.
Inversion_clear H11.
Clear H9.
Apply CLOSE2 with x.
Change (btrans (p y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Inversion_clear H6; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H11; Auto.
Inversion_clear H7; Apply notin_nu; Intros.
Cut (notin y (nu [_:name](par (p2 z _) (q2 _)))); [Intro | Auto].
Inversion_clear H11; Apply notin_nu; Intros.
Cut (notin y (par (p2 z z0) (q2 z0))); [Intro | Auto].
Inversion_clear H14; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Assumption.
Apply TRANS with (nu [_:name](nu [y:name](par (p2 y _) (q2 _))));
[Change (StBisim (nu [y:name](nu [_:name]([u:name][v:name](par (p2 u v) (q2 v)) y _)))
                (nu [_:name](nu [y:name]([u:name][v:name](par (p2 u v) (q2 v)) y _))));
 Apply NU_COMM |
 Apply NU_S; Intros;
 Change (StBisim (nu [y:name](par ([_:name](p2 _ x1) y) (q2 x1)))
                 (par (nu [z:name](p2 z x1)) (q2 x1))); Apply NU_EXTR
].
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu p) (nu q0)) (cons x l)); Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Cut (ftrans (p x0) (Out x x0) (q0 x0)); [Intro | Apply H; Auto].
Exists (nu [y:name](par (q0 y) (q2 y))); Split.
Apply fRES with (cons x (cons x0 l)); Intros.
Inversion_clear H8.
Inversion_clear H11.
Clear H9.
Apply COM2 with x.
Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H6; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H11; Auto.
Inversion_clear H7; Apply notin_nu; Intros.
Cut (notin y (par (q0 z) (q2 z))); [Intro | Auto].
Inversion_clear H11; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Assumption.
Apply REF.

(* OUT action case *)

Inversion_clear H.

Inversion_clear H0.
Elim (unsat (par (nu p) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (p x) (Out n n0) (p2 x)); [Intro | Apply H; Auto].
Exists (nu [y:name](par (p2 y) q)); Split.
Apply fRES with (cons x empty); Intros.
Inversion_clear H8.
Inversion_clear H9.
Clear H11.
Apply fPAR1.
Change (ftrans (p y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H6; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H11; Auto.
Inversion_clear H7; Apply notin_nu; Intros.
Cut (notin y (par (p2 z) q)); [Intro | Auto].
Inversion_clear H11; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply NU_EXTR.
Apply f_act_notin_Out; Auto.

Exists (nu [y:name](par (p y) p0)); Split.
Apply fRES with empty; Intros.
Inversion_clear H3; Clear H2.
Apply fPAR2; Assumption.
Apply NU_EXTR.

(* INPUT actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](par (p y0) q)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (par (p x0) q) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (ho_proc_exp p4 x0); Intros.
Inversion_clear H10.
Exists [u:name](par (nu [v:name](x1 v u)) q); Split.
Change (btrans (par (nu p) q) (In x) [u:name](par ([_:name](nu [v:name](x1 v _)) u) q)); Apply bPAR1.
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H14.
Inversion_clear H15.
Clear H17.
Change (btrans (p y) ([_:name](In x) y) (x1 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (par (p z) q)); [Intro | Auto].
Inversion_clear H17; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite <- H12; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H12 in H9.
Cut ([u:name][v:name](par (x1 u v) q))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Intro; Rewrite <- H10.
Change (StBisim (nu [z:name](par ([_:name](x1 _ y) z) q)) (par (nu [v:name](x1 v y)) q)); Apply NU_EXTR.
Inversion_clear H; Inversion_clear H11; Do 2 (Apply notin_nu; Intros).
Cut (notin x0 (par (p z) q)); [Intro | Auto]; Cut (notin x0 (nu (x1 z))); [Intro | Auto].
Inversion_clear H14; Inversion_clear H15; Apply notin_par; Auto.

Exists [_:name](par (nu p) (p4 _)); Split.
Apply bPAR2; Assumption.
Cut ([u:name][v:name](par (p u) (p4 v)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Intro; Rewrite <- H10.
Apply NU_EXTR.
Inversion_clear H; Do 2 (Apply notin_nu; Intros).
Cut (notin x0 (par (p z) q)); [Intro | Auto].
Inversion_clear H12; Apply notin_par; [Assumption | Elim (BTR_L1 q p4 (In x) x0 H14 H8); Intros; Inversion_clear H15; Auto].
Apply b_act_notin_In; Auto.

(* Symmetric case *)

Inversion_clear H.
Inversion_clear H0.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H; Auto].
Exists [_:name](nu [y:name](par (p2 y _) q)); Split.
Change (btrans (nu [y:name](par (p y) q)) (In x) [_:name](nu [y:name]([u:name][v:name](par (p2 u v) q) y _)));
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H7.
Inversion_clear H8.
Clear H10.
Apply bPAR1.
Change (btrans (p y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Inversion_clear H5; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H10; Assumption.
Inversion_clear H6; Do 2 (Apply notin_nu; Intros).
Cut (notin y (nu [v:name](par (p2 z v) q))); [Intro | Auto].
Inversion_clear H11.
Cut (notin y (par (p2 z z0) q)); [Intro | Auto].
Inversion_clear H11; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Change (StBisim (nu [y0:name](par ([_:name](p2 _ y) y0) q)) (par (nu [z:name](p2 z y)) q)); Apply NU_EXTR.
Apply b_act_notin_In; Auto.

Exists [_:name](nu [y:name](par (p y) (p0 _))); Split.
Change (btrans (nu [y:name](par (p y) q)) (In x) [_:name](nu [y:name]([u:name][v:name](par (p u) (p0 v)) y _))); Apply bRES with empty; Intros.
Apply bPAR2; Assumption.
Intro; Apply NU_EXTR.

(* BOUND OUTPUT actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](par (p y0) q)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (par (p x0) q) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (ho_proc_exp p4 x0); Intros.
Inversion_clear H10.
Exists [u:name](par (nu [v:name](x1 v u)) q); Split.
Change (btrans (par (nu p) q) (bOut x) [u:name](par ([_:name](nu [v:name](x1 v _)) u) q)); Apply bPAR1.
Apply bRES with (cons x0 empty); Intros.
Inversion_clear H14.
Inversion_clear H15.
Clear H17.
Change (btrans (p y) ([_:name](bOut x) y) (x1 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (par (p z) q)); [Intro | Auto].
Inversion_clear H17; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite <- H12; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H12 in H9.
Cut ([u:name][v:name](par (x1 u v) q))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Intros; Rewrite <- H10.
Change (StBisim (nu [z:name](par ([_:name](x1 _ y) z) q)) (par (nu [v:name](x1 v y)) q)); Apply NU_EXTR.
Inversion_clear H; Inversion_clear H11; Do 2 (Apply notin_nu; Intros).
Cut (notin x0 (par (p z) q)); [Intro | Auto]; Cut (notin x0 (nu (x1 z))); [Intro | Auto].
Inversion_clear H14; Inversion_clear H15; Apply notin_par; Auto.

Exists [_:name](par (nu p) (p4 _)); Split.
Apply bPAR2; Assumption.
Cut ([u:name][v:name](par (p u) (p4 v)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Intros; Rewrite <- H10.
Apply NU_EXTR.
Inversion_clear H; Do 2 (Apply notin_nu; Intros).
Cut (notin x0 (par (p z) q)); [Intro | Auto].
Inversion_clear H12; Apply notin_par; [Assumption | Elim (BTR_L1 q p4 (bOut x) x0 H14 H8); Intros; Inversion_clear H15; Auto].
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](par (p y0) q)) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (par (p x0) q) (Out x x0) (p1 x0)); [Intro | Auto].
Inversion H2.
Elim (proc_exp p3 x0); Intros.
Inversion_clear H10.
Exists [_:name](par (x1 _) q); Split.
Apply bPAR1; Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H15.
Clear H17.
Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (par (p z) q)); [Intro | Auto].
Inversion_clear H17; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H12; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H12 in H9.
Cut ([_:name](par (x1 _) q))=p1; [Intro | Apply proc_ext with x0; Auto].
Intros; Rewrite <- H10; Apply REF.
Inversion_clear H; Inversion_clear H11; Apply notin_nu; Intros.
Cut (notin x0 (par (p z) q)); [Intro | Auto]; Cut (notin x0 (x1 z)); [Intro | Auto]; Inversion_clear H13; Apply notin_par; Assumption.

Elim (unsat nil (cons x0 empty)); Intros.
Inversion_clear H10.
Inversion_clear H12.
Clear H11 H13.
Inversion_clear H.
Cut (notin x0 (par (p x1) q)); [Intro | Auto].
Inversion_clear H.
Absurd x0=x0; [Apply Sep_proc with q; [Apply Aux1 with p3 (Out x x0); [Apply f_act_isin_Out2 | Assumption] | Assumption] | Trivial].

(* Symmetric case *)

Inversion_clear H.
Inversion_clear H0.
Elim (unsat (par (nu p) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (p x0) (bOut x) (p2 x0)); [Intro | Apply H; Auto].
Exists [_:name](nu [y:name](par (p2 y _) q)); Split.
Change (btrans (nu [y:name](par (p y) q)) (bOut x) [_:name](nu [y:name]([u:name][v:name](par (p2 u v) q) y _))); Apply bRES with (cons x0 empty); Intros; Apply bPAR1.
Inversion_clear H7.
Inversion_clear H8.
Clear H10.
Change (btrans (p y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Inversion_clear H5; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H10; Assumption.
Inversion_clear H6; Do 2 (Apply notin_nu; Intros).
Cut (notin y (nu [v:name](par (p2 z v) q))); [Intro | Auto].
Inversion_clear H11.
Cut (notin y (par (p2 z z0) q)); [Intro | Auto].
Inversion_clear H11; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros.
Change (StBisim (nu [y0:name](par ([_:name](p2 _ y) y0) q)) (par (nu [z:name](p2 z y)) q)); Apply NU_EXTR.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu p) (nu p0)) (cons x l)); Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (p x0) (Out x x0) (p0 x0)); [Intro | Apply H; Auto].
Exists [_:name](par (p0 _) q); Split.
Apply OPEN with (cons x0 empty); Intros; Inversion_clear H8; Clear H10; Apply fPAR1.
Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H5; Apply notin_nu; Intros.
Cut (notin y (par (p z) q)); [Intro | Auto].
Inversion_clear H10; Assumption.
Inversion_clear H6; Apply notin_nu; Intros.
Cut (notin y (par (p0 z) q)); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.

Exists [_:name](nu [y:name](par (p y) (p0 _))); Split.
Change (btrans (nu [y:name](par (p y) q)) (bOut x) [_:name](nu [y:name]([u:name][v:name](par (p u) (p0 v)) y _))); Apply bRES with empty; Intros.
Apply bPAR2; Assumption.
Intros; Apply NU_EXTR.

Qed.

Lemma NU_SUM: (p,q:name->proc)(StBisim (nu [y:name](sum (p y) (q y))) (sum (nu p) (nu q))).

Proof. (* This lemma depends on Lemmata FTR_L3, BTR_L3, REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (sum (p x) (q x)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Exists (nu p2); Split.
Apply fSUM1; Apply fRES with l; Intros.
Change (ftrans (p y) ([_:name]tau y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H9; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply REF.
Exists (nu p2); Split.
Apply fSUM2; Apply fRES with l; Intros.
Change (ftrans (q y) ([_:name]tau y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H9; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply REF.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (sum (p x) (q x)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Exists (nu p2); Split.
Apply fSUM1; Apply fRES with l; Intros.
Inversion_clear H9.
Change (ftrans (p y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H12; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply REF.
Exists (nu p2); Split.
Apply fSUM2; Apply fRES with l; Intros.
Inversion_clear H9.
Change (ftrans (q y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H12; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply REF.
Apply f_act_notin_Out; Auto.

(* Symmetric case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Clear H3; Apply fSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H4; Assumption.
Apply f_act_notin_tau.
Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Clear H3; Apply fSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H4; Assumption.
Apply f_act_notin_tau.
Apply REF.

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Inversion_clear H3; Apply fSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H6; Assumption.
Apply f_act_notin_Out; Auto.
Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Inversion_clear H3; Apply fSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H6; Assumption.
Apply f_act_notin_Out; Auto.
Apply REF.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (sum (p x0) (q x0)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM1; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (p y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Apply REF.

Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM2; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (q y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Apply REF.
Apply b_act_notin_In; Auto.

(* Symmetric case *)

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intro; Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intro; Apply REF.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (sum (p x0) (q x0)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM1; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (p y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Apply REF.

Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM2; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (q y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (sum (p x0) (q x0)) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Exists p1; Split.
Apply bSUM1; Apply OPEN with l; Intros.
Change (ftrans (p y) (Out x y) (p1 y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.

Exists p1; Split.
Apply bSUM2; Apply OPEN with l; Intros.
Change (ftrans (q y) (Out x y) (p1 y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.


(* Symmetric case *)

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Apply OPEN with l; Intros.
Apply fSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intros; Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Apply OPEN with l; Intros.
Apply fSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intros; Apply REF.

Qed.

Lemma NU_TAU_PREF: (p:name->proc)(StBisim (nu [x:name](tau_pref (p x))) (tau_pref (nu p))).

Proof. (* This lemma depends on Lemmata REF, L6_Light, NU_S. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [x:name](tau_pref (p x))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (tau_pref (p x)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.
Exists (nu p); Split.
Apply TAU.
Apply NU_S; Intros.
Apply L6_Light with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (tau_pref (p z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Rewrite H7; Apply REF.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [x:name](tau_pref (p x))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (tau_pref (p x)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Apply f_act_notin_Out; Auto.

(* Symmetric case *)

Inversion H.
Exists (nu p); Split.
Apply fRES with empty; Intros.
Clear H5 H6.
Apply TAU.
Apply REF.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [x0:name](tau_pref (p x0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (tau_pref (p x0)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.

Inversion_clear H.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [x0:name](tau_pref (p x0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (tau_pref (p x0)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [x0:name](tau_pref (p x0))) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (tau_pref (p x0)) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.

Inversion_clear H.

Qed.

Lemma NU_OUT_PREF: (p:name->proc)(x,y:name)
                   (StBisim (nu [z:name](out_pref x y (p z))) (out_pref x y (nu p))).

Proof. (* This lemma depends on Lemmata REF, L6_Light, NU_S. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (out_pref x y (p x0)) tau (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (out_pref x y (p x0)) (Out n n0) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H4.
Exists (nu p); Split.
Apply OUT.
Apply NU_S; Intros.
Apply L6_Light with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (out_pref x y (p z))); [Intro | Auto].
Inversion_clear H15; Assumption.
Rewrite H12; Apply REF.
Apply f_act_notin_Out; Auto.

Inversion_clear H.
Exists (nu p); Split.
Apply fRES with empty; Intros; Apply OUT.
Apply REF.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x y (p x1)) (In x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.

Inversion_clear H.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x y (p x1)) (bOut x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu p1)) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (out_pref x y (p x1)) (Out x0 x1) (p1 x1)); [Intro | Apply H0; Auto].
Inversion H2.
Absurd y=x1; Auto.
Inversion_clear H.
Cut (notin x1 (out_pref x y (p x0))); [Intro | Auto].
Inversion_clear H; Auto.

Inversion_clear H.

Qed.

Lemma NU_IN_PREF: (p:name->name->proc)(x:name)
                  (StBisim (nu [y:name](in_pref x (p y))) (in_pref x [z:name](nu [y:name](p y z)))).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (in_pref x (p x0)) tau (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (in_pref x (p x0)) (Out n n0) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Apply f_act_notin_Out; Auto.

Inversion_clear H.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x (p x1)) (In x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion H2.
Exists [z:name](nu [y:name](p y z)); Split.
Apply IN.
Cut p=p2; [Intro | Apply ho_proc_ext with x1; Auto].
Intro; Rewrite <- H5; Apply REF.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x1 (in_pref x (p z))); [Intro | Auto]; Apply notin_nu; Intros.
Inversion_clear H10; Auto.
Apply b_act_notin_In; Auto.

Inversion H.
Exists ([z:name](nu [y:name](p y z))); Split.
Apply bRES with empty; Intros.
Apply IN.
Intro; Apply REF.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x (p x1)) (bOut x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu p1)) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (in_pref x (p x1)) (Out x0 x1) (p1 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.

Inversion_clear H.

Qed.

End restriction_laws1.

Section monoidal_par.

Lemma ID_PAR: (p:proc)(StBisim (par p nil) p).

Proof. (* This lemma is proved from scratch: it is a coinductive proof. *)

Cofix; Intro; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists p3; Split; [Assumption | Apply ID_PAR].
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H1.
Inversion_clear H1.
Inversion_clear H1.
Exists (par q1 nil); Split; [Apply fPAR1; Assumption | Apply ID_PAR].

Inversion_clear H.
Exists p3; Split; [Assumption | Intro; Apply ID_PAR].
Inversion_clear H0.
Exists [_:name](par (q1 _) nil); Split; [Apply bPAR1; Assumption | Intro; Apply ID_PAR].

Inversion_clear H.
Exists p3; Split; [Assumption | Intros; Apply ID_PAR].
Inversion_clear H0.
Exists [_:name](par (q1 _) nil); Split; [Apply bPAR1; Assumption | Intros; Apply ID_PAR].

Qed.

Lemma COMM_PAR: (p,q:proc)(StBisim (par p q) (par q p)).

Proof. (* This lemma depends on Lemma NU_S: it is a coinductive proof. *)

Cofix.
Intros.
Apply sb.
Split.
Intro.
Split.
Intros.
Inversion_clear H.

Exists (par q p3).
Split.
Apply fPAR2.
Assumption.
Apply COMM_PAR.

Exists (par p3 p).
Split.
Apply fPAR1.
Assumption.
Apply COMM_PAR.

Exists (par q2 (q1 y)).
Split.
Apply COM2 with x.
Assumption.
Assumption.
Apply COMM_PAR.

Exists (par (q2 y) q1).
Split.
Apply COM1 with x.
Assumption.
Assumption.
Apply COMM_PAR.

Exists (nu [z:name](par (q2 z) (q1 z))).
Split.
Apply CLOSE2 with x.
Assumption.
Assumption.

Apply NU_S.Intros.
Apply COMM_PAR.

Exists (nu [z:name](par (q2 z) (q1 z))).
Split.
Apply CLOSE1 with x.
Assumption.
Assumption.
Apply NU_S.Intros.
Apply COMM_PAR.

Intros.
Inversion_clear H.
Exists (par p p0).
Split.
Apply fPAR2.
Assumption.
Apply COMM_PAR.

Exists (par p0 q).
Split.
Apply fPAR1.
Assumption.
Apply COMM_PAR.

Exists (par q2 (q0 y)).
Split.
Apply COM2 with x.
Assumption.
Assumption.
Apply COMM_PAR.

Exists (par (q2 y) q0).
Split.
Apply COM1 with x.
Assumption.
Assumption.
Apply COMM_PAR.

Exists (nu [z:name](par (q2 z) (q0 z))).
Split.
Apply CLOSE2 with x.
Assumption.
Assumption.
Apply NU_S.Intros.
Apply COMM_PAR.

Exists (nu [z:name](par (q2 z) (q0 z))).
Split.
Apply CLOSE1 with x.
Assumption.
Assumption.
Apply NU_S.Intros.
Apply COMM_PAR.

Split; Intros; Split; Intros.
Inversion_clear H.
Exists [_:name](par q (p3 _)).
Split.
Apply bPAR2.
Assumption.
Intro.
Apply COMM_PAR.

Exists [_:name](par (p3 _) p).
Split.
Apply bPAR1.
Assumption.
Intro.
Apply COMM_PAR.

Inversion_clear H.
Exists [_:name](par p (p0 _)).
Split.
Apply bPAR2.
Assumption.
Intro.
Apply COMM_PAR.

Exists [_:name](par (p0 _) q).
Split.
Apply bPAR1.
Assumption.
Intro.
Apply COMM_PAR.

Inversion_clear H.
Exists [_:name](par q (p3 _)).
Split.
Apply bPAR2.
Assumption.
Intros.
Apply COMM_PAR.

Exists [_:name](par (p3 _) p).
Split.
Apply bPAR1.
Assumption.
Intros.
Apply COMM_PAR.

Inversion_clear H.
Exists [_:name](par p (p0 _)).
Split.
Apply bPAR2.
Assumption.
Intros.
Apply COMM_PAR.

Exists [_:name](par (p0 _) q).
Split.
Apply bPAR1.
Assumption.
Intros.
Apply COMM_PAR.

Qed.

Lemma ASS_PAR: (p,q,r:proc)(StBisim (par (par p q) r) (par p (par q r))).

Proof. (* This lemma depends on SYM, TRANS, NU_S, NU_EXTR, COMM_PAR:
        * it is a coinductive proof.
        *)

Cofix; Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Inversion_clear H.
Inversion_clear H0.

Exists (par p4 (par q r)); Split; [Apply fPAR1; Assumption | Apply ASS_PAR].

Exists (par p (par p4 r)); Split; [Apply fPAR2; Apply fPAR1; Assumption | Apply ASS_PAR].

Exists (par (q1 y) (par q2 r)); Split; [Apply COM1 with x; [Assumption | Apply fPAR1; Assumption] | Apply ASS_PAR].

Exists (par q1 (par (q2 y) r)); Split; [Change (ftrans (par p (par q r)) tau (par q1 ([_:name](par (q2 _) r) y))); Apply COM2 with x; [Assumption | Apply bPAR1; Assumption] | Apply ASS_PAR].

Exists (nu [_:name](par (q1 _) (par (q2 _) r))); Split.
Change (ftrans (par p (par q r)) tau (nu [_:name](par (q1 _) ([u:name](par (q2 u) r) _)))); Apply CLOSE1 with x; [Assumption | Apply bPAR1; Assumption].
Apply SYM; Apply TRANS with (nu [z:name](par (par (q1 z) (q2 z)) r)); [Apply NU_S; Intros; Apply SYM; Apply ASS_PAR | Change (StBisim (nu [z:name](par ([_:name](par (q1 _) (q2 _)) z) r)) (par (nu [z:name](par (q1 z) (q2 z))) r)); Apply NU_EXTR].

Exists (nu [_:name](par (q1 _) (par (q2 _) r))); Split.
Change (ftrans (par p (par q r)) tau (nu [_:name](par (q1 _) ([u:name](par (q2 u) r) _)))); Apply CLOSE2 with x; [Assumption | Apply bPAR1; Assumption].
Apply SYM; Apply TRANS with (nu [z:name](par (par (q1 z) (q2 z)) r)); [Apply NU_S; Intros; Apply SYM; Apply ASS_PAR | Change (StBisim (nu [z:name](par ([_:name](par (q1 _) (q2 _)) z) r)) (par (nu [z:name](par (q1 z) (q2 z))) r)); Apply NU_EXTR].

Exists (par p (par q p3)); Split; [Apply fPAR2; Apply fPAR2; Assumption | Apply ASS_PAR].

Inversion_clear H0.

Exists (par (p3 y) (par q q2)); Split; [Apply COM1 with x; [Assumption | Apply fPAR2; Assumption] | Apply ASS_PAR].

Exists (par p (par (p3 y) q2)); Split; [Apply fPAR2; Apply COM1 with x; Assumption | Apply ASS_PAR].

Inversion_clear H0.

Exists (par p3 (par q (q2 y))); Split; [Change (ftrans (par p (par q r)) tau (par p3 ([_:name](par q (q2 _)) y))); Apply COM2 with x; [Assumption | Apply bPAR2; Assumption] | Apply ASS_PAR].

Exists (par p (par p3 (q2 y))); Split; [Apply fPAR2; Apply COM2 with x; Assumption | Apply ASS_PAR].

Inversion_clear H0.

Exists (nu [_:name](par (p3 _) (par q (q2 _)))); Split.
Change (ftrans (par p (par q r)) tau (nu [_:name](par (p3 _) ([u:name](par q (q2 u)) _)))); Apply CLOSE1 with x; [Assumption | Apply bPAR2; Assumption].
Apply NU_S; Intros; Apply ASS_PAR.

Exists (par p (nu [_:name](par (p3 _) (q2 _)))); Split; [Apply fPAR2; Apply CLOSE1 with x; Assumption | Apply TRANS with (nu [z:name](par p (par (p3 z) (q2 z))))].
Apply NU_S; Intros; Apply ASS_PAR.
Apply TRANS with (nu [z:name](par (par (p3 z) (q2 z)) p)); [Apply NU_S; Intros; Apply COMM_PAR | Idtac].
Apply TRANS with (par (nu [_:name](par (p3 _) (q2 _))) p); [Change (StBisim (nu [z:name](par ([_:name](par (p3 _) (q2 _)) z) p)) (par (nu [_:name](par (p3 _) (q2 _))) p)); Apply NU_EXTR | Apply COMM_PAR].

Inversion_clear H0.

Exists (nu [_:name](par (p3 _) (par q (q2 _)))); Split.
Change (ftrans (par p (par q r)) tau (nu [_:name](par (p3 _) ([u:name](par q (q2 u)) _)))); Apply CLOSE2 with x; [Assumption | Apply bPAR2; Assumption].
Apply NU_S; Intros; Apply ASS_PAR.

Exists (par p (nu [_:name](par (p3 _) (q2 _)))); Split; [Apply fPAR2; Apply CLOSE2 with x; Assumption | Apply TRANS with (nu [z:name](par p (par (p3 z) (q2 z))))].
Apply NU_S; Intros; Apply ASS_PAR.
Apply TRANS with (nu [z:name](par (par (p3 z) (q2 z)) p)); [Apply NU_S; Intros; Apply COMM_PAR | Idtac].
Apply TRANS with (par (nu [_:name](par (p3 _) (q2 _))) p); [Change (StBisim (nu [z:name](par ([_:name](par (p3 _) (q2 _)) z) p)) (par (nu [_:name](par (p3 _) (q2 _))) p)); Apply NU_EXTR | Apply COMM_PAR].

(* Symmetric case *)

Inversion_clear H.

Exists (par (par p0 q) r); Split; [Apply fPAR1; Apply fPAR1; Assumption | Apply ASS_PAR].

Inversion_clear H0.

Exists (par (par p p3) r); Split; [Apply fPAR1; Apply fPAR2; Assumption | Apply ASS_PAR].

Exists (par (par p q) p3); Split; [Apply fPAR2; Assumption | Apply ASS_PAR].

Exists (par (par p (q0 y)) q2); Split; [Change (ftrans (par (par p q) r) tau (par ([_:name](par p (q0 _)) y) q2)); Apply COM1 with x; [Apply bPAR2; Assumption | Assumption]| Apply ASS_PAR].

Exists (par (par p q0) (q2 y)); Split; [Apply COM2 with x; [Apply fPAR2; Assumption | Assumption]| Apply ASS_PAR].

Exists (nu [_:name](par (par p (q0 _)) (q2 _))); Split; [Change (ftrans (par (par p q) r) tau (nu [_:name](par ([u:name](par p (q0 u)) _) (q2 _)))); Apply CLOSE1 with x; [Apply bPAR2; Assumption | Assumption] | Idtac].
Apply TRANS with (nu [_:name](par p (par (q0 _) (q2 _)))).
Apply NU_S; Intros; Apply ASS_PAR.
Apply TRANS with (nu [_:name](par (par (q0 _) (q2 _)) p)).
Apply NU_S; Intros; Apply COMM_PAR.
Apply TRANS with (par (nu [z:name](par (q0 z) (q2 z))) p); [Change (StBisim (nu [_:name](par ([u:name](par (q0 u) (q2 u)) _) p)) (par (nu [z:name](par (q0 z) (q2 z))) p)); Apply NU_EXTR | Apply COMM_PAR].

Exists (nu [_:name](par (par p (q0 _)) (q2 _))); Split; [Change (ftrans (par (par p q) r) tau (nu [_:name](par ([u:name](par p (q0 u)) _) (q2 _)))); Apply CLOSE2 with x; [Apply bPAR2; Assumption | Assumption] | Idtac].
Apply TRANS with (nu [_:name](par p (par (q0 _) (q2 _)))).
Apply NU_S; Intros; Apply ASS_PAR.
Apply TRANS with (nu [_:name](par (par (q0 _) (q2 _)) p)).
Apply NU_S; Intros; Apply COMM_PAR.
Apply TRANS with (par (nu [z:name](par (q0 z) (q2 z))) p); [Change (StBisim (nu [_:name](par ([u:name](par (q0 u) (q2 u)) _) p)) (par (nu [z:name](par (q0 z) (q2 z))) p)); Apply NU_EXTR | Apply COMM_PAR].

Inversion_clear H1.

Exists (par (par (q0 y) p0) r); Split; [Apply fPAR1; Apply COM1 with x; Assumption | Apply ASS_PAR].

Exists (par (par (q0 y) q) p0); Split; [Change (ftrans (par (par p q) r) tau (par ([_:name](par (q0 _) q) y) p0)); Apply COM1 with x; [Apply bPAR1; Assumption | Assumption] | Apply ASS_PAR].

Inversion_clear H1.

Exists (par (par q0 (p0 y)) r); Split; [Apply fPAR1; Apply COM2 with x; Assumption | Apply ASS_PAR].

Exists (par (par q0 q) (p0 y)); Split; [Apply COM2 with x; [Apply fPAR1; Assumption | Assumption] | Apply ASS_PAR].

Inversion_clear H1.

Exists (par (nu [_:name](par (q0 _) (p0 _))) r); Split.
Apply fPAR1; Apply CLOSE1 with x; Assumption.
Apply SYM; Apply TRANS with (nu [z:name](par (par (q0 z) (p0 z)) r)).
Apply NU_S; Intros; Apply SYM; Apply ASS_PAR.
Change (StBisim (nu [z:name](par ([_:name](par (q0 _) (p0 _)) z) r)) (par (nu [_:name](par (q0 _) (p0 _))) r)); Apply NU_EXTR.

Exists (nu [_:name](par (par (q0 _) q) (p0 _))); Split; [Change (ftrans (par (par p q) r) tau (nu [_:name](par ([u:name](par (q0 u) q) _) (p0 _)))); Apply CLOSE1 with x; [Apply bPAR1; Assumption | Assumption] | Idtac].
Apply NU_S; Intros; Apply ASS_PAR.

Inversion_clear H1.

Exists (par (nu [_:name](par (q0 _) (p0 _))) r); Split.
Apply fPAR1; Apply CLOSE2 with x; Assumption.
Apply SYM; Apply TRANS with (nu [z:name](par (par (q0 z) (p0 z)) r)).
Apply NU_S; Intros; Apply SYM; Apply ASS_PAR.
Change (StBisim (nu [z:name](par ([_:name](par (q0 _) (p0 _)) z) r)) (par (nu [_:name](par (q0 _) (p0 _))) r)); Apply NU_EXTR.

Exists (nu [_:name](par (par (q0 _) q) (p0 _))); Split; [Change (ftrans (par (par p q) r) tau (nu [_:name](par ([u:name](par (q0 u) q) _) (p0 _)))); Apply CLOSE2 with x; [Apply bPAR1; Assumption | Assumption] | Idtac].
Apply NU_S; Intros; Apply ASS_PAR.

(* INPUT actions case *)

Inversion_clear H.
Inversion_clear H0.

Exists [_:name](par (p4 _) (par q r)); Split; [Apply bPAR1; Assumption | Intro; Apply ASS_PAR].

Exists [_:name](par p (par (p4 _) r)); Split; [Change (btrans (par p (par q r)) (In x) [_:name](par p ([u:name](par (p4 u) r) _))); Apply bPAR2; Apply bPAR1; Assumption | Intro; Apply ASS_PAR].

Exists [_:name](par p (par q (p3 _))); Split; [Change (btrans (par p (par q r)) (In x) [_:name](par p ([u:name](par q (p3 u)) _))); Apply bPAR2; Apply bPAR2; Assumption | Intro; Apply ASS_PAR].

(* Symmetric case *)

Inversion_clear H.

Exists [_:name](par (par (p0 _) q) r); Split; [Change (btrans (par (par p q) r) (In x) [_:name](par ([u:name](par (p0 u) q) _) r)); Apply bPAR1; Apply bPAR1; Assumption | Intro; Apply ASS_PAR].

Inversion_clear H0.

Exists [_:name](par (par p (p3 _)) r); Split; [Change (btrans (par (par p q) r) (In x) [_:name](par ([u:name](par p (p3 u)) _) r)); Apply bPAR1; Apply bPAR2; Assumption | Intro; Apply ASS_PAR].

Exists [_:name](par (par p q) (p3 _)); Split; [Apply bPAR2; Assumption | Intro; Apply ASS_PAR].

(* BOUND OUTPUT actions case *)

Inversion_clear H.
Inversion_clear H0.

Exists [_:name](par (p4 _) (par q r)); Split; [Apply bPAR1; Assumption | Intros; Apply ASS_PAR].

Exists [_:name](par p (par (p4 _) r)); Split; [Change (btrans (par p (par q r)) (bOut x) [_:name](par p ([u:name](par (p4 u) r) _))); Apply bPAR2; Apply bPAR1; Assumption | Intros; Apply ASS_PAR].

Exists [_:name](par p (par q (p3 _))); Split; [Change (btrans (par p (par q r)) (bOut x) [_:name](par p ([u:name](par q (p3 u)) _))); Apply bPAR2; Apply bPAR2; Assumption | Intros; Apply ASS_PAR].

(* Symmetric case *)

Inversion_clear H.

Exists [_:name](par (par (p0 _) q) r); Split; [Change (btrans (par (par p q) r) (bOut x) [_:name](par ([u:name](par (p0 u) q) _) r)); Apply bPAR1; Apply bPAR1; Assumption | Intros; Apply ASS_PAR].

Inversion_clear H0.

Exists [_:name](par (par p (p3 _)) r); Split; [Change (btrans (par (par p q) r) (bOut x) [_:name](par ([u:name](par p (p3 u)) _) r)); Apply bPAR1; Apply bPAR2; Assumption | Intros; Apply ASS_PAR].

Exists [_:name](par (par p q) (p3 _)); Split; [Apply bPAR2; Assumption | Intros; Apply ASS_PAR].

Qed.

End monoidal_par.

Section structural_congruence2.

Lemma PAR_S2: (p,q,r,s:proc)(StBisim p q)->(StBisim r s)->(StBisim (par p r) (par q s)).

Proof. (* This lemma depends on Lemmata TRANS, PAR_S and COMM_PAR. *)

(Intros; Apply TRANS with (par s q);
 [ Apply TRANS with (par r q);
   [ Apply TRANS with (par q r);
     [ (Apply PAR_S; Assumption) | Apply COMM_PAR ]
   | (Apply PAR_S; Assumption) ]
 | Apply COMM_PAR ]).

Qed.

Transparent PAR_S2.

End structural_congruence2.

Section restriction_laws2.

Lemma NU_P: (p:proc)(StBisim (nu [x:name]p) p).

Proof. (* This lemma depends on Lemmata REF, SYM, TRANS, NU_S,
        * ID_PAR, COMM_PAR, NU_EXTR, PAR_S2, NU_NIL.
        *)

Intro; Apply TRANS with (nu [_:name](par p nil)).
Apply NU_S; Intros; Apply SYM; Apply ID_PAR.
Apply TRANS with (nu [_:name](par nil p)).
Apply NU_S; Intros; Apply COMM_PAR.
Apply TRANS with (par nil p).
Apply TRANS with (par (nu [x:name]nil) p).
Change (StBisim (nu [_:name](par ([x:name]nil _) p)) (par (nu [x:name]nil) p)); Apply NU_EXTR.
Apply PAR_S2; [Apply NU_NIL | Apply REF].
Apply TRANS with (par p nil); [Apply COMM_PAR | Apply ID_PAR].

Qed.

Lemma NU_PAR1: (p:name->proc)(q:proc)(StBisim (nu [x:name](par (p x) q)) (par (nu p) (nu [x:name]q))).

Proof. (* This lemma depends on Lemmata REF, SYM, TRANS, NU_EXTR, NU_P, PAR_S2. *)

Intros; Apply TRANS with (par (nu p) q); [Apply NU_EXTR | Apply PAR_S2; [Apply REF | Apply SYM; Apply NU_P]].

Qed.

Lemma NU_PAR2: (p:proc)(q:name->proc)(StBisim (nu [x:name](par p (q x))) (par (nu [x:name]p) (nu q))).

Proof. (* This lemma depends on Lemmata REF, SYM, TRANS, NU_EXTR, NU_P, PAR_S2, COMM_PAR. *)

Intros; Apply TRANS with (nu [x:name](par (q x) p)); [Apply NU_S; Intros; Apply COMM_PAR | Apply TRANS with (par (nu q) p); [Apply NU_EXTR | Apply TRANS with (par (nu q) (nu [_:name]p)); [Apply PAR_S2; [Apply REF | Apply SYM; Apply NU_P] | Apply COMM_PAR]]].

Qed.

End restriction_laws2.

Section matching_laws.

Lemma MATCH1: (p:proc)(x,y:name)(x=y)->(StBisim (match x y p) p).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H0.
Exists p1; Split; [Assumption | Apply REF].
Exists q1; Split; [Rewrite H; Apply fMATCH; Assumption | Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intro; Apply REF].
Exists q1; Split; [Rewrite H; Apply bMATCH; Assumption | Intro; Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intros; Apply REF].
Exists q1; Split; [Rewrite H; Apply bMATCH; Assumption | Intros; Apply REF].

Qed.

Lemma MATCH2: (p:proc)(x,y:name)~(x=y)->(StBisim (match x y p) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Qed.

Lemma MISMATCH1: (p:proc)(x,y:name)~(x=y)->(StBisim (mismatch x y p) p).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H0.
Exists p1; Split; [Assumption | Apply REF].
Exists q1; Split; [Apply fMISMATCH; Assumption | Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intro; Apply REF].
Exists q1; Split; [Apply bMISMATCH; Assumption | Intro; Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intros; Apply REF].
Exists q1; Split; [Apply bMISMATCH; Assumption | Intros; Apply REF].

Qed.

Lemma MISMATCH2: (p:proc)(x,y:name)(x=y)->(StBisim (mismatch x y p) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Qed.

End matching_laws.
